"""
Generate Amber topology files from GROMACS files
"""

# ##############################################################################
#                           GPLv3 LICENSE INFO                                 #
#                                                                              #
#  Copyright (C) 2020  Mario S. Valdes-Tresanco and Mario E. Valdes-Tresanco   #
#  Copyright (C) 2014  Jason Swails, Bill Miller III, and Dwight McGee         #
#                                                                              #
#   Project: https://github.com/Valdes-Tresanco-MS/gmx_MMPBSA                  #
#                                                                              #
#   This program is free software; you can redistribute it and/or modify it    #
#  under the terms of the GNU General Public License version 3 as published    #
#  by the Free Software Foundation.                                            #
#                                                                              #
#  This program is distributed in the hope that it will be useful, but         #
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    #
#  for more details.                                                           #
# ##############################################################################

import os
import parmed
from GMXMMPBSA.exceptions import *
from GMXMMPBSA.utils import checkff
from GMXMMPBSA.alamdcrd import _scaledistance
import subprocess
from math import sqrt
from pathlib import Path
import logging
import string
from parmed.tools.changeradii import ChRad

chains_letters = list(string.ascii_uppercase)
his = ['HIS', 'HIE', 'HID', 'HIP']
cys_name = ['CYS', 'CYX', 'CYM']
lys = ['LYS', 'LYN']
asp = ['ASP', 'ASH']
glu = ['GLU', 'GLH']
PBRadii = {1: 'bondi', 2: 'mbondi', 3: 'mbondi2', 4: 'mbondi3'}

ions_para_files = {1: 'frcmod.ions234lm_126_tip3p', 2: 'frcmod.ions234lm_iod_tip4pew', 3: 'frcmod.ions234lm_iod_spce',
                   4: 'frcmod.ions234lm_hfe_spce', 5: 'frcmod.ions234lm_126_tip4pew', 6: 'frcmod.ions234lm_126_spce',
                   7: 'frcmod.ions234lm_1264_tip4pew', 8: 'frcmod.ions234lm_1264_tip3p',
                   9: 'frcmod.ions234lm_1264_spce', 10: 'frcmod.ions234lm_iod_tip3p',
                   11: 'frcmod.ions234lm_hfe_tip4pew', 12: 'frcmod.ions234lm_hfe_tip3p}'}

ions = ["AG", "AL", "Ag", "BA", "BR", "Be", "CA", "CD", "CE", "CL", "CO", "CR", "CS", "CU", "CU1", "Ce", "Cl-", "Cr",
        "Dy", "EU", "EU3", "Er", "F", "FE", "FE2", "GD3", "H3O+", "HE+", "HG", "HZ+", "Hf", "IN", "IOD", "K", "K+",
        "LA", "LI", "LU", "MG", "MN", "NA", "NH4", "NI", "Na+", "Nd", "PB", "PD", "PR", "PT", "Pu", "RB", "Ra", "SM",
        "SR", "Sm", "Sn", "TB", "TL", "Th", "Tl", "Tm", "U4+", "V2+", "Y", "YB2", "ZN", "Zr"]


def dist(coor1, coor2):
    return sqrt((coor2[0] - coor1[0]) ** 2 + (coor2[1] - coor1[1]) ** 2 + (coor2[2] - coor1[2]) ** 2)


class CheckMakeTop:
    def __init__(self, FILES, INPUT, external_programs):
        self.FILES = FILES
        self.INPUT = INPUT
        self.external_progs = external_programs
        self.use_temp = False
        self.log = open('gmx_MMPBSA.log', 'a')
        self.print_residues = 'within' in self.INPUT['print_res']  # FIXME: this is pretty ugly
        self.within = 4
        if self.print_residues:
            self.within = float(self.INPUT['print_res'].split()[1])

        # Define Gromacs executable
        self.make_ndx = self.external_progs['make_ndx']
        self.trjconv = self.external_progs['trjconv']
        self.editconf = self.external_progs['editconf']

        self.ref_str = None

        self.ligand_tpr = None
        self.ligand_mol2 = None

        self.rec_str_ions = False
        self.lig_str_ions = False

        # create the * prmtop variables for compatibility with the original code
        self.complex_pmrtop = 'COM.prmtop'
        self.receptor_pmrtop = 'REC.prmtop'
        self.ligand_pmrtop = 'LIG.prmtop'

        self.mutant_complex_pmrtop = 'MUT_COM.prmtop'
        self.mutant_receptor_pmrtop = 'MUT_REC.prmtop'
        self.mutant_ligand_pmrtop = 'MUT_LIG.prmtop'

        self.complex_temp_top = self.FILES.prefix + 'COM.top'
        self.receptor_temp_top = self.FILES.prefix + 'COM.top'
        self.ligand_temp_top = self.FILES.prefix + 'COM.top'

        self.complex_str_file = self.FILES.prefix + 'COM.pdb'
        self.receptor_str_file = self.FILES.prefix + 'REC.pdb'
        self.ligand_str_file = self.FILES.prefix + 'LIG.pdb'

        if self.FILES.reference_structure:
            self.ref_str = parmed.read_PDB(self.FILES.reference_structure)

        checkff(self.INPUT['overwrite_data'])
        self.checkFiles()

    def checkFiles(self):
        if (not self.FILES.complex_tpr or not self.FILES.complex_index or not self.FILES.complex_trajs or not
        self.FILES.complex_groups):
            GMXMMPBSA_ERROR('You must define the structure, topology and index files, as well as the groups!')

    def buildTopology(self):
        """
        :return: complex, receptor, ligand topologies and their mutants
        """
        self.gmx2pdb()
        if self.FILES.complex_top:
            logging.info('Cleaning GROMACS topologies...')
            self.cleantop(self.FILES.complex_top, self.complex_temp_top)
            if self.FILES.receptor_top:
                self.cleantop(self.FILES.receptor_top, self.receptor_temp_top)
            if self.FILES.ligand_top:
                self.cleantop(self.FILES.ligand_top, self.ligand_temp_top)
            tops = self.gmxtop2prmtop()
        else:
            self.pdb2prmtop()
            tops = self.makeToptleap()

        self.reswithin()
        self.cleanup_trajs()
        return tops

    def gmx2pdb(self):
        """
        Generate PDB file to generate topology
        :return:
        """

        logging.info('Get PDB files from GROMACS structures files...')

        # wt complex
        # make index for extract pdb structure
        rec_group, lig_group = self.FILES.complex_groups

        logging.info('Making gmx_MMPBSA index for complex...')
        # merge both (rec and lig) groups into complex group, modify index and create a copy
        # 1-rename groups, 2-merge
        make_ndx_echo_args = ['echo', 'name {r} GMXMMPBSA_REC\n name {l} GMXMMPBSA_LIG\n  {r} | '
                                      '{l}\n q\n'.format(r=rec_group, l=lig_group)]
        c1 = subprocess.Popen(make_ndx_echo_args, stdout=subprocess.PIPE)

        com_ndx = self.FILES.prefix + 'COM_index.ndx'
        make_ndx_args = self.make_ndx + ['-n', self.FILES.complex_index, '-o', com_ndx]
        if self.INPUT['debug_printlevel']:
            logging.info('Running command: ' + (' '.join(make_ndx_echo_args).replace('\n', '\\n')) + ' | ' +
                         ' '.join(make_ndx_args))
        c2 = subprocess.Popen(make_ndx_args, stdin=c1.stdout, stdout=self.log, stderr=self.log)
        if c2.wait():  # if it quits with return code != 0
            GMXMMPBSA_ERROR('%s failed when querying %s' % (' '.join(self.make_ndx), self.FILES.complex_index))
        self.FILES.complex_index = com_ndx

        logging.info('Normal Complex: Saving group {}_{} in {} file as '
                     '{}'.format(rec_group, lig_group, self.FILES.complex_index, self.complex_str_file))
        editconf_echo_args = ['echo', 'GMXMMPBSA_REC_GMXMMPBSA_LIG']
        c3 = subprocess.Popen(editconf_echo_args, stdout=subprocess.PIPE)
        # we extract a pdb from structure file to make amber topology
        editconf_args = self.editconf + ['-f', self.FILES.complex_tpr, '-o', self.complex_str_file, '-n',
                                    self.FILES.complex_index]
        if self.INPUT['debug_printlevel']:
            logging.info('Running command: ' + (' '.join(editconf_echo_args)) + ' | ' + ' '.join(editconf_args))
        c4 = subprocess.Popen(editconf_args, stdin=c3.stdout, stdout=self.log, stderr=self.log)
        if c4.wait():  # if it quits with return code != 0
            GMXMMPBSA_ERROR('%s failed when querying %s' % (' '.join(self.editconf), self.FILES.complex_tpr))

        # Put receptor and ligand (explicitly defined) to avoid overwrite them
        # check if ligand is not protein. In any case, non-protein ligand always most be processed
        if self.FILES.ligand_mol2:
            logging.info(f'Generating ligand parameters from {self.FILES.ligand_mol2} file...')
            lig_name = os.path.splitext(os.path.split(self.FILES.ligand_mol2)[1])[0]
            self.ligand_frcmod = self.FILES.prefix + lig_name + '.frcmod'
            # run parmchk2
            parmchk2 = self.external_progs['parmchk2']
            parmchk2_args = [parmchk2, '-i', self.FILES.ligand_mol2, '-f', 'mol2', '-o', self.ligand_frcmod]
            if self.INPUT['debug_printlevel']:
                logging.info('Running command: ' + ' '.join(parmchk2_args))
            l3 = subprocess.Popen(parmchk2_args, stdout=self.log, stderr=self.log)
            if l3.wait():
                GMXMMPBSA_ERROR('%s failed when querying %s' % (parmchk2, self.FILES.ligand_mol2))

        # make a temp receptor pdb (even when stability) if decomp to get correct receptor residues from complex. This
        # avoid get multiples molecules from complex.split()
        if self.INPUT['decomprun'] and self.print_residues:
            if self.FILES.stability:
                self.use_temp = True
                logging.warning('When decomp is defined, we generate a receptor file in order to extract interface '
                                'residues')
                rec_echo_args = ['echo', '{}'.format(rec_group)]
                cp1 = subprocess.Popen(rec_echo_args, stdout=subprocess.PIPE)
                # we extract a pdb from structure file to make amber topology
                editconf_args = self.editconf + ['-f', self.FILES.complex_tpr, '-o', 'rec_temp.pdb', '-n',
                                            self.FILES.complex_index]
                if self.INPUT['debug_printlevel']:
                    logging.info('Running command: ' + (' '.join(rec_echo_args)) + ' | ' + ' '.join(editconf_args))
                cp2 = subprocess.Popen(editconf_args, stdin=cp1.stdout, stdout=self.log, stderr=self.log)
                if cp2.wait():  # if it quits with return code != 0
                    GMXMMPBSA_ERROR('%s failed when querying %s' % (' '.join(self.editconf), self.FILES.complex_tpr))

        # check if stability
        if self.FILES.stability:
            if (self.FILES.receptor_tpr or self.FILES.ligand_tpr):
                logging.warning('When Stability calculation mode is selected, receptor and ligand files are not '
                                'needed...')
            if self.INPUT['alarun'] and (self.FILES.mutant_receptor_tpr or self.FILES.mutant_ligand_tpr):
                logging.warning('When Stability calculation mode is selected, mutant receptor/mutant ligand files '
                                'are not needed...')

        # wt receptor
        if self.FILES.receptor_tpr:
            logging.info('A receptor structure file was defined. Using MT approach...')
            logging.info('Normal receptor: Saving group {} in {} file as {}'.format(
                self.FILES.receptor_group, self.FILES.receptor_index, self.receptor_str_file))
            editconf_echo_args = ['echo', '{}'.format(self.FILES.receptor_group)]
            p1 = subprocess.Popen(editconf_echo_args, stdout=subprocess.PIPE)
            # we extract a pdb from structure file to make amber topology
            editconf_args = self.editconf + ['-f', self.FILES.receptor_tpr, '-o', self.receptor_str_file, '-n',
                                        self.FILES.receptor_index]
            if self.INPUT['debug_printlevel']:
                logging.info('Running command: ' + (' '.join(editconf_echo_args)) + ' | ' + ' '.join(editconf_args))
            cp2 = subprocess.Popen(editconf_args, stdin=p1.stdout, stdout=self.log, stderr=self.log)
            if cp2.wait():  # if it quits with return code != 0
                GMXMMPBSA_ERROR('%s failed when querying %s' % (' '.join(self.editconf), self.FILES.receptor_tpr))

        else:
            logging.info('No receptor structure file was defined. Using ST approach...')
            logging.info('Using receptor structure from complex to generate AMBER topology')
            # wt complex receptor
            logging.info('Normal Complex: Saving group {} in {} file as {}'.format(
                rec_group, self.FILES.complex_index, self.receptor_str_file))
            editconf_echo_args = ['echo', '{}'.format(rec_group)]
            cp1 = subprocess.Popen(editconf_echo_args, stdout=subprocess.PIPE)
            # we extract a pdb from structure file to make amber topology
            editconf_args = self.editconf + ['-f', self.FILES.complex_tpr, '-o', self.receptor_str_file, '-n',
                                        self.FILES.complex_index]
            if self.INPUT['debug_printlevel']:
                logging.info('Running command: ' + (' '.join(editconf_echo_args)) + ' | ' + ' '.join(editconf_args))
            cp2 = subprocess.Popen(editconf_args, stdin=cp1.stdout, stdout=self.log, stderr=self.log)
            if cp2.wait():  # if it quits with return code != 0
                GMXMMPBSA_ERROR('%s failed when querying %s' % (' '.join(self.editconf), self.FILES.complex_tpr))

        # ligand
        # # check consistence
        if self.FILES.ligand_tpr:  # ligand is protein
            # FIXME: if ligand is a zwitterionic aa fail
            logging.info('A ligand structure file was defined. Using MT approach...')
            logging.info('Normal Ligand: Saving group {} in {} file as {}'.format(
                self.FILES.ligand_group, self.FILES.ligand_index, self.ligand_str_file))
            # wt ligand
            editconf_echo_args = ['echo', '{}'.format(self.FILES.ligand_group)]
            l1 = subprocess.Popen(editconf_echo_args, stdout=subprocess.PIPE)
            # we extract a pdb from structure file to make amber topology
            editconf_args = self.editconf + ['-f', self.FILES.ligand_tpr, '-o', self.ligand_str_file, '-n',
                                        self.FILES.ligand_index]
            if self.INPUT['debug_printlevel']:
                logging.info('Running command: ' + (' '.join(editconf_echo_args)) + ' | ' + ' '.join(editconf_args))
            l2 = subprocess.Popen(editconf_args, stdin=l1.stdout, stdout=self.log, stderr=self.log)
            if l2.wait():  # if it quits with return code != 0
                GMXMMPBSA_ERROR('%s failed when querying %s' % (' '.join(self.editconf), self.FILES.ligand_tpr))
        else:
            # wt complex ligand
            logging.info('No ligand structure file was defined. Using ST approach...')
            logging.info('Using ligand structure from complex to generate AMBER topology')
            logging.info('Normal ligand: Saving group {} in {} file as {}'.format(lig_group, self.FILES.complex_index,
                                                                                                  self.ligand_str_file))
            editconf_echo_args = ['echo', '{}'.format(lig_group)]
            cl1 = subprocess.Popen(editconf_echo_args, stdout=subprocess.PIPE)
            # we extract a pdb from structure file to make amber topology
            editconf_args = self.editconf + ['-f', self.FILES.complex_tpr, '-o', self.ligand_str_file, '-n',
                                        self.FILES.complex_index]
            if self.INPUT['debug_printlevel']:
                logging.info('Running command: ' + (' '.join(editconf_echo_args)) + ' | ' + ' '.join(editconf_args))
            cl2 = subprocess.Popen(editconf_args, stdin=cl1.stdout, stdout=self.log, stderr=self.log)
            if cl2.wait():  # if it quits with return code != 0
                GMXMMPBSA_ERROR('%s failed when querying %s' % (' '.join(self.editconf), self.FILES.complex_tpr))

        # initialize receptor and ligand structures. Needed to get residues map
        self.complex_str = self.molstr(self.complex_str_file)
        self.receptor_str = self.molstr(self.receptor_str_file)
        self.ligand_str = self.molstr(self.ligand_str_file)
        self.resi, self.resl, self.orderl = self.res2map()
        self.fix_chains_IDs(self.complex_str, self.receptor_str, self.ligand_str, self.ref_str)

    def gmxtop2prmtop(self):
        logging.info('Building Normal Complex Amber Topology...')
        com_top = parmed.gromacs.GromacsTopologyFile(self.complex_temp_top, xyz=self.complex_str_file)
        # try:
        if com_top.impropers or com_top.urey_bradleys or com_top.cmaps:
            com_amb_prm = parmed.amber.ChamberParm.from_structure(com_top)
            com_top_parm = 'chamber'
        else:
            com_amb_prm = parmed.amber.AmberParm.from_structure(com_top)
            com_top_parm = 'amber'

        # IMPORTANT: make_trajs ends in error if the box is defined
        com_amb_prm.box = None
        # except TypeError as err:
        #     GMXMMPBSA_ERROR(str(err))

        self.fixparm2amber(com_amb_prm)

        logging.info('Writing Normal Complex Amber Topology...')
        # change de PBRadii
        action = ChRad(com_amb_prm, PBRadii[self.INPUT['PBRadii']])
        com_amb_prm.write_parm(self.complex_pmrtop)

        text_list = []
        for r in self.resi['REC']:
            if r[0] == r[1]:
                text_list.append(f'{r[0]}')
            else:
                text_list.append(f'{r[0]}-{r[1]}')
        rec_indexes_string = ','.join(text_list)

        rec_hastop = True
        if self.FILES.receptor_top:
            logging.info('A Receptor topology file was defined. Using MT approach...')
            logging.info('Building AMBER Receptor Topology from GROMACS Receptor Topology...')
            rec_top = parmed.gromacs.GromacsTopologyFile(self.receptor_temp_top, xyz=self.receptor_str_file)
            if rec_top.impropers or rec_top.urey_bradleys or rec_top.cmaps:
                if com_top_parm == 'amber':
                    GMXMMPBSA_ERROR('Inconsistent parameter format. The defined Complex is AMBER type while the '
                                    'Receptor is CHAMBER type!')
                rec_amb_prm = parmed.amber.ChamberParm.from_structure(rec_top)
            else:
                if com_top_parm == 'chamber':
                    GMXMMPBSA_ERROR('Inconsistent parameter format. The defined Complex is CHAMBER type while the '
                                    'Receptor is AMBER type!')
                rec_amb_prm = parmed.amber.AmberParm.from_structure(rec_top)
            self.fixparm2amber(rec_amb_prm)
        else:
            logging.info('No Receptor topology files was defined. Using ST approach...')
            logging.info('Building AMBER Receptor Topology from Complex...')
            # we make a copy for receptor topology
            rec_amb_prm = self.molstr(com_amb_prm)
            rec_amb_prm.strip(f'!:{rec_indexes_string}')
            rec_hastop = False
        # change de PBRadii
        action = ChRad(rec_amb_prm, PBRadii[self.INPUT['PBRadii']])
        rec_amb_prm.write_parm(self.receptor_pmrtop)

        lig_hastop = True
        if self.FILES.ligand_top:
            logging.info('A Ligand Topology file was defined. Using MT approach...')
            logging.info('Building AMBER Ligand Topology from GROMACS Ligand Topology...')
            lig_top = parmed.gromacs.GromacsTopologyFile(self.ligand_temp_top, xyz=self.ligand_str_file)
            if lig_top.impropers or lig_top.urey_bradleys or lig_top.cmaps:
                if com_top_parm == 'amber':
                    GMXMMPBSA_ERROR('Inconsistent parameter format. The defined Complex is AMBER type while the '
                                    'Ligand is CHAMBER type!')
                lig_amb_prm = parmed.amber.ChamberParm.from_structure(lig_top)
            else:
                if com_top_parm == 'chamber':
                    GMXMMPBSA_ERROR('Inconsistent parameter format. The defined Complex is CHAMBER type while the '
                                    'Ligand is AMBER type!')
                lig_amb_prm = parmed.amber.AmberParm.from_structure(lig_top)
            self.fixparm2amber(lig_amb_prm)
        else:
            logging.info('No Ligand Topology files was defined. Using ST approach...')
            logging.info('Building AMBER Ligand Topology from Complex...')
            # we make a copy for ligand topology
            lig_amb_prm = self.molstr(com_amb_prm)
            lig_amb_prm.strip(f':{rec_indexes_string}')
            lig_hastop = False

        # change de PBRadii
        action = ChRad(lig_amb_prm, PBRadii[self.INPUT['PBRadii']])
        lig_amb_prm.write_parm(self.ligand_pmrtop)

        if self.INPUT['alarun']:
            logging.info('Building Mutant Complex Topology...')
            # get mutation index in complex
            com_mut_index, part_mut, part_index, mut_label = self.getMutationIndex()
            mut_com_amb_prm = self.makeMutTop(com_amb_prm, com_mut_index)
            # change de PBRadii
            action = ChRad(mut_com_amb_prm, PBRadii[self.INPUT['PBRadii']])
            mut_com_amb_prm.write_parm(self.mutant_complex_pmrtop)

            if part_mut == 'REC':
                logging.info('Detecting mutation in Receptor. Building Mutant Receptor Topology...')
                mut_com_amb_prm.strip(f'!:{rec_indexes_string}')
                mdata = [self.mutant_receptor_pmrtop, 'REC']
                self.mutant_ligand_pmrtop = None
                if rec_hastop:
                    mtop = self.makeMutTop(rec_amb_prm, part_index)
                else:
                    mtop = mut_com_amb_prm
            else:
                logging.info('Detecting mutation in Ligand. Building Mutant Ligand Topology...')
                mut_com_amb_prm.strip(f':{rec_indexes_string}')
                mdata = [self.mutant_ligand_pmrtop, 'LIG']
                self.mutant_receptor_pmrtop = None
                if lig_hastop:
                    mtop = self.makeMutTop(lig_amb_prm, part_index)
                else:
                    mtop = mut_com_amb_prm

            if self.INPUT['protein_forcefield'] == 'charmm':
                mut_prot_amb_prm = parmed.amber.ChamberParm.from_structure(mtop)
            else:
                mut_prot_amb_prm = parmed.amber.AmberParm.from_structure(mtop)
            # change de PBRadii
            action = ChRad(mut_prot_amb_prm, PBRadii[self.INPUT['PBRadii']])
            mut_prot_amb_prm.write_parm(mdata[0])
        else:
            self.mutant_complex_pmrtop = None

        return (self.complex_pmrtop, self.receptor_pmrtop, self.ligand_pmrtop, self.mutant_complex_pmrtop,
                self.mutant_receptor_pmrtop, self.mutant_ligand_pmrtop)

    def pdb2prmtop(self):
        """
        Generate parmed structure object for complex, receptor and ligand ( if it is protein-like)
        :return:
        """
        logging.info('Generating AMBER Compatible PDB Files...')

        # fix receptor and structures
        self.fixparm2amber(self.complex_str, removeH=True)
        self.fixparm2amber(self.receptor_str, removeH=True)
        self.fixparm2amber(self.ligand_str, removeH=True)

        self.receptor_list = {}
        start = 1
        c = 1
        for r in self.resi['REC']:
            end = start + (r[1]- r[0])
            mask = f'!:{start}-{end}'
            start += end
            rec = self.molstr(self.receptor_str)
            rec.strip(mask)
            rec_file = self.FILES.prefix + f'REC_F{c}.pdb'
            rec.save(rec_file, 'pdb', True, renumber=False)
            self.receptor_list[f'REC{c}'] = rec_file
            c += 1

        self.ligand_list = {}
        start = 1
        c = 1
        for r in self.resi['LIG']:
            end = start + (r[1] - r[0])
            mask = f'!:{start}-{end}'
            start += end
            lig = self.molstr(self.ligand_str)
            lig.strip(mask)
            lig_file = self.FILES.prefix + f'LIG_F{c}.pdb'
            lig.save(lig_file, 'pdb', True, renumber=False)
            self.ligand_list[f'LIG{c}'] = lig_file
            c += 1

        self.mut_receptor_list = {}
        self.mut_ligand_list = {}

        if self.INPUT['alarun']:
            com_mut_index, part_mut, part_index, mut_label = self.getMutationIndex()
            if part_mut == 'REC':
                logging.info('Detecting mutation in Receptor. Building Mutant Receptor Structure...')
                self.mutant_ligand_pmrtop = None
                start = 0
                c = 1
                for r in self.resi['REC']:
                    mask = f'!:{start}-{(r[1] - r[0]) + 1}'
                    rec = self.molstr(self.receptor_str)
                    mut_rec = self.makeMutTop(rec, part_index)
                    mut_rec.strip(mask)
                    mut_rec_file = self.FILES.prefix + f'MUT_REC_F{c}.pdb'
                    mut_rec.save(mut_rec_file, 'pdb', True, renumber=False)
                    self.mut_receptor_list[f'MREC{c}'] = mut_rec_file
                    c += 1
            else:
                logging.info('Detecting mutation in Ligand.Building Mutant Ligand Structure...')
                self.mutant_receptor_pmrtop = None
                start = 0
                c = 1
                for r in self.resi['LIG']:
                    mask = f'!:{start}-{(r[1] - r[0]) + 1}'
                    lig = self.molstr(self.ligand_str)
                    mut_lig = self.makeMutTop(lig, part_index)
                    mut_lig.strip(mask)
                    mut_lig_file = self.FILES.prefix + f'MUT_LIG_F{c}.pdb'
                    mut_lig.save(mut_lig_file, 'pdb', True, renumber=False)
                    self.mut_ligand_list[f'MLIG{c}'] = mut_lig_file
                    c += 1

    def reswithin(self):
        # Get residue form receptor-ligand interface
        if self.print_residues and self.INPUT['decomprun']:
            res_list = []
            for i in self.resl['REC']:
                if i in res_list:
                    continue
                for j in self.resl['LIG']:
                    if j in res_list:
                        continue
                    for rat in self.complex_str.residues[i - 1].atoms:
                        rat_coor = [rat.xx, rat.xy, rat.xz]
                        if i in res_list:
                            break
                        for lat in self.complex_str.residues[j - 1].atoms:
                            lat_coor = [lat.xx, lat.xy, lat.xz]
                            if dist(rat_coor, lat_coor) <= self.within:
                                if i not in res_list:
                                    res_list.append(i)
                                if j not in res_list:
                                    res_list.append(j)
                                break
            res_list.sort()
            self.INPUT['print_res'] = ','.join([str(x + 1) for x in res_list])

    def cleantop(self, top_file, temp_top_file):
        """
        Create a new top file without SOL and IONS
        :param top_file: User-defined topology file
        :param temp_top_file: temporary top file
        :return: detected ff
        """
        top_file = Path(top_file)
        molsect = False

        temp_top = open(top_file.parent.joinpath(temp_top_file), 'w')
        temp_top.write('; Modified by gmx_MMPBSA\n')

        with open(top_file) as topf:
            for line in topf:
                if line.startswith('#include') and 'forcefield.itp' in line:
                    if not ('charmm' in line.lower() or 'toppar' in line.lower() or 'amber' in line.lower()):
                        GMXMMPBSA_ERROR(f'Unknown force field in GROMACS topology in line:\n {line}')
                elif '[ molecules ]' in line:
                    molsect = True
                if molsect:
                    # not copy ions and solvent
                    sol_ion = [
                        # standard gmx form
                        'NA', 'CL', 'SOL',
                        # charmm-GUI form ??
                        'SOD', 'CLA', 'TIP3P', 'TIP4P', 'TIPS3P', 'TIP5P', 'SPC', 'SPC/E', 'SPCE', 'TIP3o', 'WAT']
                    if line.split()[0].strip().upper() in sol_ion:
                        break
                temp_top.write(line)
        temp_top.close()

    def get_masks(self):
        rt = []
        for r in self.resi['REC']:
            if r[0] == r[1]: # only as precaution
                rt.append(f'{r[0]}')
            else:
                rt.append(f'{r[0]}-{r[1]}')
        rec_mask = ':' + ','.join(rt)
        lt = []
        for l in self.resi['LIG']:
            if l[0] == l[1]:
                lt.append(f'{l[0]}')
            else:
                lt.append(f'{l[0]}-{l[1]}')
        lig_mask = ':' + ','.join(lt)
        return rec_mask, lig_mask

    def res2map(self):
        """

        :param com_str:
        :return:
        """
         # read the index file
        ndx = {}
        with open(self.FILES.complex_index) as indexf:
            header = None
            for line in indexf:
                if line.startswith('['):
                    header = line.strip('\n[] ')
                    ndx[header] = []
                else:
                    ndx[header].extend(map(int, line.split()))
        com_str = self.complex_str

        start = 1
        end = None
        order_list = []
        previous = None

        masks = {'REC': [], 'LIG': []}
        res_list = {'REC': [], 'LIG': []}
        com_ndx = ndx['GMXMMPBSA_REC_GMXMMPBSA_LIG']
        com_len = len(ndx['GMXMMPBSA_REC_GMXMMPBSA_LIG'])
        resnum = 1
        current_res = None
        for i in range(com_len):
            # We check who owns the residue corresponding to this atom
            if com_ndx[i] in ndx['GMXMMPBSA_REC']:
                current = 'R'
                # save residue number in the rec list
                if com_str.atoms[i].residue.number != current_res and not resnum in res_list['REC']:
                    res_list['REC'].append(resnum)
                    resnum += 1
                    current_res = com_str.atoms[i].residue.number
            else:
                current = 'L'
                # save residue number in the lig list
                if com_str.atoms[i].residue.number != current_res and not resnum in res_list['LIG']:
                    res_list['LIG'].append(resnum)
                    resnum += 1
                    current_res = com_str.atoms[i].residue.number
            # check for end
            if previous and current != previous:
                end = resnum - 2

            # when i is the last index
            if i == com_len - 1:
                end = resnum - 1
            if end:
                if previous == 'R':
                    masks['REC'].append([start, end])
                else:
                    masks['LIG'].append([start, end])
                # add current range identifier
                order_list.append(previous)
                # set the new start and reset end
                start = end + 1
                end = None
            # we change previous to current once it is processed
            previous = current
        return masks, res_list, order_list

    def fixparm2amber(self, structure, removeH=False):

        for residue in structure.residues:
            # change atoms name from GROMACS to AMBER
            if residue.name == 'ILE':
                for atom in residue.atoms:
                    if atom.name == 'CD':
                        atom.name = 'CD1'
            for atom in residue.atoms:
                if atom.name == 'OC1':
                    atom.name = 'O'
                elif atom.name == 'OC2':
                    atom.name = 'OXT'
                    residue.ter = True  # parmed terminal
            # change residues name according to AMBER
            if residue.name == 'LYS':
                atoms = [atom.name for atom in residue.atoms]
                if not 'HZ3' in atoms:
                    residue.name = 'LYN'
            elif residue.name == 'ASP':
                atoms = [atom.name for atom in residue.atoms]
                if 'HD2' in atoms:
                    residue.name = 'ASH'
            elif residue.name == 'GLU':
                atoms = [atom.name for atom in residue.atoms]
                if 'HE2' in atoms:
                    residue.name = 'GLH'
            elif residue.name in his:
                atoms = [atom.name for atom in residue.atoms if atom.atomic_number == 1]
                if 'HD1' in atoms and 'HE2' in atoms:
                    residue.name = 'HIP'
                elif 'HD1' in atoms:
                    residue.name = 'HID'
                elif 'HE2' in atoms:
                    residue.name = 'HIE'
            elif residue.name in cys_name:
                if residue.name == 'CYX':
                    continue
                for atom in residue.atoms:
                    if 'SG' in atom.name:
                        for bondedatm in atom.bond_partners:
                            if bondedatm.name == 'SG':
                                residue.name = 'CYX'
                                bondedatm.residue.name = 'CYX'
                        break
            # GROMACS 4.x save the pdb without atom element column, so parmed does not recognize some H atoms.
            # Parmed assigns 0 to the atomic number of these atoms. In order to correctly eliminate hydrogens,
            # it is necessary to assign the atomic number.
            if len(self.make_ndx) == 2:
                for atom in residue.atoms:
                    if 'H' in atom.name and atom.atomic_number == 0:
                        atom.atomic_number = 1
            # Remove H atoms. Only when using the pdb files with tleap to build the topologies
        if removeH:
            structure.strip('@/H')

    def getMutationIndex(self):
        label = ''
        if not self.INPUT['mutant_res']:
            GMXMMPBSA_ERROR("No residue for mutation was defined")
        chain_, resnum_ = self.INPUT['mutant_res'].split(':')
        chain = str(chain_).strip().upper()
        resnum = int(str(resnum_).strip())

        if not chain or not resnum:
            GMXMMPBSA_ERROR("Wrong notation... You most define the residue to mutate as follow: CHAIN:RES_NUMBER")
        idx = 0
        for res in self.complex_str.residues:
            if res.number == int(resnum) and res.chain == chain:
                try:
                    parmed.residue.AminoAcidResidue.get(res.name, True)
                except KeyError as e:
                    GMXMMPBSA_ERROR('The mutation must be an amino acid residue ...')

                label = f'{res.chain}:{res.name}:{res.number}'
                break
            idx += 1

        if idx in self.resl['REC']:
            part_index = self.resl['REC'].index(idx)
            part_mut = 'REC'
        elif idx in self.resl['LIG']:
            part_index = self.resl['LIG'].index(idx)
            part_mut = 'LIG'
        else:
            part_index = None
            part_mut = None
            GMXMMPBSA_ERROR('Residue {}:{} not found'.format(chain, resnum))

        return (idx, part_mut, part_index, label)

    def makeMutTop(self, wt_top, mut_index):
        """

        :param wt_top: Amber parm from GROMACS topology
        :param mut_index: index of mutation in structure
        :return: Mutant AmberParm
        """
        mut_top = self.molstr(wt_top)
        mut_aa = self.INPUT['mutant']

        bb_atoms = 'N,H,CA,HA,C,O,'
        nterm_atoms = 'H1,H2,H3,'
        cterm_atoms = 'OXT'
        sc_cb_atom = 'CB,'
        sc_ala_atoms = ('HB,' +  # VAL, ILE, THR
                        'HB1,HB2,' +
                        'CG1,CG2,OG1,' +  # VAL, ILE, THR
                        'OG,' +  # SER
                        'SG,' +  # CYS
                        'CG,')

        if mut_aa in ['GLY', 'G']:
            # FIXME: allow terminal residues to mutate?
            strip_mask = f':{mut_index + 1} &!@' + bb_atoms + sc_cb_atom + nterm_atoms + cterm_atoms
        else:
            # FIXME: allow terminal residues to mutate?
            strip_mask = f':{mut_index + 1} &!@' + bb_atoms + sc_cb_atom + sc_ala_atoms + nterm_atoms + cterm_atoms
        mut_top.strip(strip_mask)

        h_atoms_prop = {}
        for res in mut_top.residues:
            if res.name == mut_aa:
                for at in res.atoms:
                    if mut_aa == 'GLY':
                        if at.name in ['HA2']:
                            h_atoms_prop['mass'] = at.mass
                            h_atoms_prop['element'] = at.element
                            h_atoms_prop['atomic_number'] = at.atomic_number
                            h_atoms_prop['charge'] = at.charge
                            h_atoms_prop['atom_type'] = at.atom_type
                            h_atoms_prop['type'] = at.type
                            break
                    else:
                        if at.name in ['HB2']:
                            h_atoms_prop['mass'] = at.mass
                            h_atoms_prop['element'] = at.element
                            h_atoms_prop['atomic_number'] = at.atomic_number
                            h_atoms_prop['charge'] = at.charge
                            h_atoms_prop['atom_type'] = at.atom_type
                            h_atoms_prop['type'] = at.type
                            break
                break

        cb_atom = None
        ca_atom = None
        mut_top.residues[mut_index].name = mut_aa
        ind = 0
        for res in mut_top.residues:
            if ind == mut_index:
                for at in res.atoms:
                    if mut_aa == 'GlY':
                        if at.name == 'CA':
                            ca_atom = at
                        if at.name in ['CB']:
                            at.name = 'HA2'
                            ca_atom.xx, ca_atom.xy, ca_atom.xz, at.xx, at.xy, at.xz = _scaledistance(
                                [ca_atom.xx, ca_atom.xy,
                                 ca_atom.xz, at.xx, at.xy,
                                 at.xz], 1.09)
                            for ref_at in h_atoms_prop:
                                setattr(at, ref_at, h_atoms_prop[ref_at])
                        elif at.name in ['HA']:
                            at.name = 'HA1'
                            at.type = h_atoms_prop['type']
                            at.atom_type = h_atoms_prop['atom_type']
                    else:
                        if at.name == 'CB':
                            cb_atom = at
                        if at.name in ['CG', 'OG', 'CG2', 'SG']:
                            at.name = 'HB3'
                            cb_atom.xx, cb_atom.xy, cb_atom.xz, at.xx, at.xy, at.xz = _scaledistance(
                                [cb_atom.xx, cb_atom.xy,
                                 cb_atom.xz, at.xx, at.xy,
                                 at.xz], 1.09)
                            for ref_at in h_atoms_prop:
                                setattr(at, ref_at, h_atoms_prop[ref_at])
                        elif at.name in ['HB']:
                            at.name = 'HB1'
                            at.type = h_atoms_prop['type']
                            at.atom_type = h_atoms_prop['atom_type']
                        elif at.name in ['HB1', 'HB2', 'HB3']:
                            at.type = h_atoms_prop['type']
                            at.atom_type = h_atoms_prop['atom_type']
            ind += 1
        return mut_top

    def cleanup_trajs(self):
        # clear trajectory
        if self.INPUT['solvated_trajectory']:
            logging.info('Cleaning normal complex trajectories...')
            new_trajs = []
            for i in range(len(self.FILES.complex_trajs)):
                trjconv_echo_args = ['echo', 'GMXMMPBSA_REC_GMXMMPBSA_LIG']
                c5 = subprocess.Popen(trjconv_echo_args, stdout=subprocess.PIPE)
                # we get only first trajectory to extract a pdb file and make amber topology for complex
                trjconv_args = self.trjconv + ['-f', self.FILES.complex_trajs[0], '-s', self.FILES.complex_tpr,
                                          '-o', 'COM_traj_{}.xtc'.format(i), '-n', self.FILES.complex_index]
                if self.INPUT['debug_printlevel']:
                    logging.info('Running command: ' + (' '.join(trjconv_echo_args)) + ' | ' + ' '.join(trjconv_args))
                c6 = subprocess.Popen(trjconv_args,  # FIXME: start and end frames???
                                      stdin=c5.stdout, stdout=self.log, stderr=self.log)
                if c6.wait():  # if it quits with return code != 0
                    GMXMMPBSA_ERROR('%s failed when querying %s' % (' '.join(self.trjconv), self.FILES.complex_tpr))
                new_trajs.append('COM_traj_{}.xtc'.format(i))
            self.FILES.complex_trajs = new_trajs

            # clear trajectory
            if self.FILES.receptor_tpr:
                logging.info('Cleaning normal receptor trajectories...')
                new_trajs = []
                for i in range(len(self.FILES.receptor_trajs)):
                    trjconv_echo_args = ['echo', '{}'.format(self.FILES.receptor_group)]
                    c5 = subprocess.Popen(trjconv_echo_args, stdout=subprocess.PIPE)
                    # we get only first trajectory to extract a pdb file and make amber topology for complex
                    trjconv_args = self.trjconv + ['-f', self.FILES.receptor_trajs[0], '-s', self.FILES.receptor_tpr,
                                              '-o', 'REC_traj_{}.xtc'.format(i), '-n',
                                              self.FILES.receptor_index]
                    if self.INPUT['debug_printlevel']:
                        logging.info('Running command: ' + (' '.join(trjconv_echo_args)) + ' | ' + ' '.join(
                            trjconv_args))
                    c6 = subprocess.Popen(trjconv_args,  # FIXME: start and end frames???
                                          stdin=c5.stdout, stdout=self.log, stderr=self.log)
                    if c6.wait():  # if it quits with return code != 0
                        GMXMMPBSA_ERROR(
                            '%s failed when querying %s' % (' '.join(self.trjconv), self.FILES.receptor_tpr))
                    new_trajs.append('REC_traj_{}.xtc'.format(i))
                self.FILES.receptor_trajs = new_trajs

            if self.FILES.ligand_tpr:
                logging.info('Cleanig normal ligand trajectories...')
                new_trajs = []
                for i in range(len(self.FILES.ligand_trajs)):
                    trjconv_echo_args = ['echo', '{}'.format(self.FILES.ligand_group)]
                    c5 = subprocess.Popen(trjconv_echo_args, stdout=subprocess.PIPE)
                    # we get only first trajectory to extract a pdb file and make amber topology for complex
                    trjconv_args = self.trjconv + ['-f', self.FILES.ligand_trajs[0], '-s', self.FILES.ligand_tpr,
                                              '-o', 'LIG_traj_{}.xtc'.format(i), '-n', self.FILES.ligand_index]
                    if self.INPUT['debug_printlevel']:
                        logging.info(
                            'Running command: ' + (' '.join(trjconv_echo_args)) + ' | ' + ' '.join(trjconv_args))
                    c6 = subprocess.Popen(trjconv_args, stdin=c5.stdout, stdout=self.log, stderr=self.log)
                    if c6.wait():  # if it quits with return code != 0
                        GMXMMPBSA_ERROR('%s failed when querying %s' % (' '.join(self.trjconv), self.FILES.ligand_tpr))
                    new_trajs.append('LIG_traj_{}.xtc'.format(i))
                self.FILES.ligand_trajs = new_trajs

    def fix_chains_IDs(self, com_str, rec_str=None, lig_str=None, ref_str=None):
        if ref_str:
            if len(ref_str.residues) != len(com_str.residues):
                GMXMMPBSA_ERROR('The number of amino acids in the reference structure is different from that of the '
                                'complex...')
            c = 0
            for res in ref_str.residues:
                res.chain = com_str.residues[c].chain
                if c + 1 in self.resl['REC']:
                    i = self.resl['REC'].index(c)
                    rec_str.residues[i].chain = res.chain
                else:
                    i = self.resl['LIG'].index(c)
                    lig_str.residues[i].chain = res.chain
                c += 1
        else:
            assign = False
            if self.INPUT['assign_chainID'] == 1:
                assign = not com_str.residues[0].chain  # pretty simple
                if assign:
                    logging.info('Chains ID not found. Assigning chains IDs...')
                else:
                    logging.info('Chains ID found. Ignoring chains ID assignation...')
            elif self.INPUT['assign_chainID'] == 2:
                assign = True
                if not com_str.residues[0].chain:
                    logging.warning('Already have chain ID. Re-assigning ID...')
                else:
                    logging.warning('Assigning chains ID...')
            elif self.INPUT['assign_chainID'] == 0 and self.FILES.complex_tpr[-3] == 'gro':
                assign = True
                logging.warning('No reference structure was found and a gro file was used for the complex '
                                'structure. Assigning chains ID...')

            if assign:
                chains_ids = []
                chain_by_num = False
                chain_by_ter = False
                previous_res_number = 0
                curr_chain_id = 'A'
                has_nucl = 0
                c = 0
                for res in com_str.residues:
                    if not res.chain:
                        res.chain = curr_chain_id

                        if c + 1 in self.resl['REC']:
                            i = self.resl['REC'].index(c)
                            rec_str.residues[i].chain = res.chain
                        else:
                            i = self.resl['LIG'].index(c)
                            lig_str.residues[i].chain = res.chain
                        if curr_chain_id not in chains_ids:
                            chains_ids.append(curr_chain_id)
                    else:
                        if res.chain != curr_chain_id:
                            res.chain = curr_chain_id
                            if c in self.resl['REC']:
                                i = self.resl['REC'].index(c)
                                rec_str.residues[i].chain = res.chain
                            else:
                                i = self.resl['LIG'].index(c)
                                lig_str.residues[i].chain = res.chain
                        if res.chain not in chains_ids:
                            chains_ids.append(res.chain)
                    # see if it is the end of chain
                    if res.number != previous_res_number + 1:
                        if previous_res_number != 0:
                            chain_by_num = True
                    if chain_by_num and chain_by_ter:
                        chain_by_num = False
                        chain_by_ter = False
                        curr_chain_id = chains_letters[chains_letters.index(chains_ids[-1]) + 1]
                        res.chain = curr_chain_id
                        if c in self.resl['REC']:
                            i = self.resl['REC'].index(c)
                            rec_str.residues[i].chain = res.chain
                        else:
                            i = self.resl['LIG'].index(c)
                            lig_str.residues[i].chain = res.chain
                        if res.chain not in chains_ids:
                            chains_ids.append(res.chain)
                    elif chain_by_ter:
                        chain_by_ter = False
                    elif chain_by_num:
                        chain_by_num = False
                        curr_chain_id = chains_letters[chains_letters.index(chains_ids[-1]) + 1]
                        res.chain = curr_chain_id

                        if c + 1 in self.resl['REC']:
                            i = self.resl['REC'].index(c)
                            rec_str.residues[i].chain = res.chain
                        else:
                            i = self.resl['LIG'].index(c)
                            lig_str.residues[i].chain = res.chain
                        if res.chain not in chains_ids:
                            chains_ids.append(res.chain)
                    for atm in res.atoms:
                        if atm.name == 'OXT':  # only for protein
                            res.ter = True
                            chain_by_ter = True
                    if parmed.residue.RNAResidue.has(res.name) or parmed.residue.DNAResidue.has(res.name):
                        has_nucl += 1

                    previous_res_number = res.number
                    c += 1
                if has_nucl == 1:
                    logging.warning('This structure contains nucleotides. We recommend that you use the reference '
                                    'structure')

    def molstr(self, data):

        if type(data) == str:
            # data is a pdb file
            pdb_file = data
            try:
                with open(pdb_file) as fo:
                    fo = fo.readlines()
                    for line in fo:
                        if 'MODEL' in line or 'ENDMDL' in line:
                            fo.remove(line)
                with open(pdb_file, 'w') as fw:
                    for x in fo:
                        fw.write(x)
            except IOError as e:
                GMXMMPBSA_ERROR('', str(e))

            structure = parmed.read_PDB(pdb_file)
        else:
            # data is Structure, AmberParm, ChamberParm or GromacsTopologyFile. This make a copy
            structure = data.__copy__()
            c = 1
            for at in structure.atoms:
                at.number = c
                c += 1
        return structure

    def makeToptleap(self):
        logging.info('Building Tleap input files...')
        with open(self.FILES.prefix + 'leap.in', 'w') as tif:
            tif.write('source {}\n'.format(self.INPUT['protein_forcefield']))
            tif.write('source {}\n'.format(self.INPUT['ligand_forcefield']))
            tif.write('loadOff atomic_ions.lib\n')
            tif.write('loadamberparams {}\n'.format(ions_para_files[self.INPUT['ions_parameters']]))
            tif.write('set default PBRadii {}\n'.format(PBRadii[self.INPUT['PBRadii']]))

            REC = []
            LIG = []
            for rec in self.receptor_list:
                REC.append(f'{rec}')
                tif.write(f'{rec} = loadpdb {self.receptor_list[rec]}\n')
            rec_out = ' '.join(REC)

            # check if ligand is not protein and always load
            if self.FILES.ligand_mol2:
                tif.write('LIG1 = loadmol2 {}\n'.format(self.FILES.ligand_mol2))
                tif.write('check LIG1\n')
                tif.write('loadamberparams {}\n'.format(self.ligand_frcmod))
                if not self.FILES.stability:
                    tif.write('saveamberparm LIG1 {t} {p}LIG.inpcrd\n'.format(t=self.ligand_pmrtop,
                                                                              p=self.FILES.prefix))
                else:
                    self.ligand_pmrtop = None
                for lig in self.ligand_list:
                    LIG.append(f'{lig}')
            else:
                for lig in self.ligand_list:
                    LIG.append(f'{lig}')
                    tif.write(f'{lig} = loadpdb {self.ligand_list[lig]}\n')
                lig_out = ' '.join(LIG)
                if not self.FILES.stability:
                    tif.write(f'LIG_OUT = combine {{ {lig_out} }}\n')
                    tif.write('saveamberparm LIG_OUT {t} {p}LIG.inpcrd\n'.format(t=self.ligand_pmrtop,
                                                                                 p=self.FILES.prefix))
                else:
                    self.ligand_pmrtop = None
            COM = []
            l = 0
            r = 0
            i = 0
            for e in self.orderl:
                if e == 'R':
                    COM.append(REC[r])
                    r += 1
                else:
                    COM.append(LIG[l])
                    l += 1
                i += 1

            if not self.FILES.stability:
                tif.write(f'REC_OUT = combine {{ { rec_out } }}\n')
                tif.write('saveamberparm REC_OUT {t} {p}REC.inpcrd\n'.format(t=self.receptor_pmrtop, p=self.FILES.prefix))
            else:
                self.receptor_pmrtop = None
            com_out = ' '.join(COM)
            tif.write(f'COM_OUT = combine {{ {com_out} }}\n')
            tif.write('saveamberparm COM_OUT {t} {p}COM.inpcrd\n'.format(t=self.complex_pmrtop, p=self.FILES.prefix))
            tif.write('quit')

        tleap = self.external_progs['tleap']
        tleap_args = [tleap, '-f', '{}'.format(self.FILES.prefix + 'leap.in')]
        if self.INPUT['debug_printlevel']:
            logging.info('Running command: ' + ' '.join(tleap_args))
        p = subprocess.Popen(tleap_args, stdout=self.log, stderr=self.log)
        if p.wait():
            GMXMMPBSA_ERROR('%s failed when querying %s' % (tleap, self.FILES.prefix + 'leap.in'))

        if self.INPUT['alarun']:
            with open(self.FILES.prefix + 'mut_leap.in', 'w') as mtif:
                mtif.write('source {}\n'.format(self.INPUT['protein_forcefield']))
                mtif.write('source {}\n'.format(self.INPUT['ligand_forcefield']))
                mtif.write('loadOff atomic_ions.lib\n')
                mtif.write('loadamberparams {}\n'.format(ions_para_files[self.INPUT['ions_parameters']]))
                mtif.write('set default PBRadii {}\n'.format(PBRadii[self.INPUT['PBRadii']]))


                if self.mutant_receptor_pmrtop:
                    REC = []
                    for mrec in self.mut_receptor_list:
                        REC.append(f'{mrec}')
                        mtif.write(f'{mrec} = loadpdb {self.mut_receptor_list[mrec]}\n')
                    mrec_out = ' '.join(REC)

                    if not self.FILES.stability:
                        mtif.write(f'MREC_OUT = combine {{ {mrec_out} }}\n')
                        mtif.write('saveamberparm MREC_OUT {t} {p}MUT_REC.inpcrd\n'.format(t=self.mutant_receptor_pmrtop,
                                                                                         p=self.FILES.prefix))
                    else:
                        self.mutant_receptor_pmrtop = None
                    # check if ligand is not protein and always load
                    if self.FILES.ligand_mol2:
                        mtif.write('LIG1 = loadmol2 {}\n'.format(self.FILES.ligand_mol2))
                        self.mutant_ligand_pmrtop = None
                        if not self.FILES.stability:
                            mtif.write('check LIG1\n')
                            mtif.write('loadamberparams {}\n'.format(self.ligand_frcmod))
                        else:
                            self.mutant_ligand_pmrtop = None
                    else:
                        for lig in self.ligand_list:
                            mtif.write(f'{lig} = loadpdb {self.ligand_list[lig]}\n')
                else:
                    LIG = []
                    for mlig in self.mut_ligand_list:
                        LIG.append(f'{mlig}')
                        mtif.write(f'{mlig} = loadpdb {self.mut_ligand_list[lig]}\n')
                    mlig_out = ' '.join(LIG)

                    if not self.FILES.stability:
                        mtif.write(f'MLIG_OUT = combine {{ {mlig_out} }}\n')
                        mtif.write('saveamberparm MLIG_OUT {t} {p}MUT_LIG.inpcrd\n'.format(
                                t=self.mutant_ligand_pmrtop, p=self.FILES.prefix))
                    else:
                        self.mutant_ligand_pmrtop = None
                    for rec in self.receptor_list:
                        mtif.write(f'{rec} = loadpdb {self.receptor_list[rec]}\n')

                MCOM = []
                l = r = i = 0
                for e in self.orderl:
                    if e == 'R':
                        MCOM.append(REC[r])
                        r += 1
                    else:
                        MCOM.append(LIG[l])
                        l += 1
                    i += 1
                mcom_out = ' '.join(MCOM)
                mtif.write(f'MCOM_OUT = combine {{ {mcom_out} }}\n')
                mtif.write('saveamberparm MCOM_OUT {t} {p}MUT_COM.inpcrd\n'.format(t=self.mutant_complex_pmrtop,
                                                                               p=self.FILES.prefix))
                mtif.write('quit')

            tleap_args = [tleap, '-f', '{}'.format(self.FILES.prefix + 'mut_leap.in')]
            if self.INPUT['debug_printlevel']:
                logging.info('Running command: ' + ' '.join(tleap_args))
            p1 = subprocess.Popen(tleap_args, stdout=self.log, stderr=self.log)
            if p1.wait():
                GMXMMPBSA_ERROR('%s failed when querying %s' % (tleap, self.FILES.prefix + 'mut_leap.in'))
        else:
            self.mutant_complex_pmrtop = None

        return (self.complex_pmrtop, self.receptor_pmrtop, self.ligand_pmrtop, self.mutant_complex_pmrtop,
                self.mutant_receptor_pmrtop, self.mutant_ligand_pmrtop)



"""
This module contains classes for creating input files for various MM/PBSA
calculations.

Methods:
   create_inputs -- determines which input files need to be written then 
                    handles it

Classes: 
    NabInput -- base class of mmpbsa_py_energy input files
    GBNabInput -- writes input files for GB (derived from NabInput)
    PBNabInput -- writes input files for PB (derived from NabInput)
    SanderInput -- Base class for sander input files
    SanderGBInput -- writes input files for sander GB (derived from SanderInput)
    SanderPBSAInput- writes input files for sander PB (derived from SanderInput)
    SanderAPBSInput -- writes input files for sander.APBS PB 
                       (derived from SanderInput)
    SanderGBDecomp -- writes GB decomp input files (derived from SanderInput)
    SanderPBDecomp -- writes PB decomp input files (derived from SanderInput)
    QuasiHarmonicInput -- writes a cpptraj input file for quasi-harmonic calcs
"""

# ##############################################################################
#                           GPLv3 LICENSE INFO                                 #
#                                                                              #
#  Copyright (C) 2020  Mario S. Valdes-Tresanco and Mario E. Valdes-Tresanco   #
#  Copyright (C) 2014  Jason Swails, Bill Miller III, and Dwight McGee         #
#                                                                              #
#   Project: https://github.com/Valdes-Tresanco-MS/gmx_MMPBSA                  #
#                                                                              #
#   This program is free software; you can redistribute it and/or modify it    #
#  under the terms of the GNU General Public License version 3 as published    #
#  by the Free Software Foundation.                                            #
#                                                                              #
#  This program is distributed in the hope that it will be useful, but         #
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    #
#  for more details.                                                           #
# ##############################################################################

from __future__ import division

from parmed.amber.mdin import Mdin
from parmed.exceptions import AmberError
from copy import deepcopy


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

def create_inputs(INPUT, prmtop_system, pre):
    """ Creates the input files for all necessary calculations """
    stability = prmtop_system.stability

    # First check if we are running decomp
    if INPUT['decomprun']:
        # Get the cards that will go in the complex mdin file
        com_card, rc_card, lc_card = prmtop_system.Group(INPUT['print_res'], True)
        junk, rec_card, lig_card = prmtop_system.Group(INPUT['print_res'], False,
                                                       rec_group_type='RES', lig_group_type='RES')
        full_com, full_rc, full_lc = prmtop_system.Group('all', True)
        junk, full_rec, full_lig = prmtop_system.Group('all', False)

        # Convert the strings into card objects
        # Now create the mdin objects
        if INPUT['gbrun']:
            if stability:
                rec_res = ['Residues considered as REC', full_rc]
                pri_res = ['Residues to print', com_card]
                com_mdin = SanderGBDecomp(INPUT, rec_res, pri_res)
                com_mdin.write_input(pre + 'gb_decomp_com.mdin')
            else:
                rec_res = ['Residues considered as REC', full_rc]
                lig_res = ['Residues considered as LIG', full_lc]
                pri_res = ['Residues to print', com_card]
                com_mdin = SanderGBDecomp(INPUT, rec_res, lig_res, pri_res)
                rec_res = ['Residues considered as REC', full_rec]
                pri_res = ['Residues to print', rec_card]
                rec_mdin = SanderGBDecomp(INPUT, rec_res, pri_res)
                lig_res = ['Residues considered as LIG', full_lig]
                pri_res = ['Residues to print', lig_card]
                lig_mdin = SanderGBDecomp(INPUT, lig_res, pri_res)
                com_mdin.write_input(pre + 'gb_decomp_com.mdin')
                rec_mdin.write_input(pre + 'gb_decomp_rec.mdin')
                lig_mdin.write_input(pre + 'gb_decomp_lig.mdin')
        if INPUT['pbrun']:
            if stability:
                rec_res = ['Residues considered as REC', full_rc]
                pri_res = ['Residues to print', com_card]
                com_mdin = SanderPBDecomp(INPUT, rec_res, pri_res)
                com_mdin.write_input(pre + 'pb_decomp_com.mdin')
            else:
                rec_res = ['Residues considered as REC', full_rc]
                lig_res = ['Residues considered as LIG', full_lc]
                pri_res = ['Residues to print', com_card]
                com_mdin = SanderPBDecomp(INPUT, rec_res, lig_res, pri_res)
                rec_res = ['Residues considered as REC', full_rec]
                pri_res = ['Residues to print', rec_card]
                rec_mdin = SanderPBDecomp(INPUT, rec_res, pri_res)
                lig_res = ['Residues considered as LIG', full_lig]
                pri_res = ['Residues to print', lig_card]
                lig_mdin = SanderPBDecomp(INPUT, lig_res, pri_res)
                com_mdin.write_input(pre + 'pb_decomp_com.mdin')
                rec_mdin.write_input(pre + 'pb_decomp_rec.mdin')
                lig_mdin.write_input(pre + 'pb_decomp_lig.mdin')

    else:  # not decomp

        if INPUT['gbrun']:
            gb_prog = 'mmpbsa_py_energy'
            if INPUT['use_sander'] or INPUT['ifqnt'] == 1:
                gb_prog = 'sander'

            if gb_prog == 'mmpbsa_py_energy':
                gb_mdin = GBNabInput(INPUT)
                gb_mdin.write_input(pre + 'gb.mdin')
            else:
                # We need separate input files for QM/gmx_MMPBSA
                if INPUT['ifqnt']:
                    com_input = deepcopy(INPUT)
                    rec_input = deepcopy(INPUT)
                    lig_input = deepcopy(INPUT)
                    (com_input['qmmask'], rec_input['qmmask'],
                     lig_input['qmmask']) = prmtop_system.Mask(INPUT['qm_residues'],
                                                               in_complex=False)
                    if not com_input['qmmask']:
                        raise AmberError('No valid QM residues chosen!')
                    com_input['qm_theory'] = "'%s'" % com_input['qm_theory']
                    com_input['qmmask'] = "'%s'" % com_input['qmmask']
                    com_input['qmcharge'] = com_input['qmcharge_com']
                    gb_mdin = SanderGBInput(com_input)
                    gb_mdin.write_input(pre + 'gb_qmmm_com.mdin')
                    if not stability:
                        if not rec_input['qmmask']:
                            rec_input['ifqnt'] = 0
                        else:
                            rec_input['qmmask'] = "'%s'" % rec_input['qmmask']
                        rec_input['qm_theory'] = "'%s'" % rec_input['qm_theory']
                        rec_input['qmcharge'] = rec_input['qmcharge_rec']
                        gb_mdin = SanderGBInput(rec_input)
                        gb_mdin.write_input(pre + 'gb_qmmm_rec.mdin')
                        if not lig_input['qmmask']:
                            lig_input['ifqnt'] = 0
                        else:
                            lig_input['qmmask'] = "'%s'" % lig_input['qmmask']
                        lig_input['qm_theory'] = "'%s'" % lig_input['qm_theory']
                        lig_input['qmcharge'] = lig_input['qmcharge_lig']
                        gb_mdin = SanderGBInput(lig_input)
                        gb_mdin.write_input(pre + 'gb_qmmm_lig.mdin')
                else:
                    gb_mdin = SanderGBInput(INPUT)
                    gb_mdin.write_input(pre + 'gb.mdin')

        if INPUT['pbrun']:
            pb_prog = 'mmpbsa_py_energy'
            if INPUT['sander_apbs']:
                pb_prog = 'sander.APBS'
            elif INPUT['use_sander']:
                pb_prog = 'sander'

            if pb_prog == 'mmpbsa_py_energy':
                pb_mdin = PBNabInput(INPUT)
                pb_mdin2 = PBNabInput(INPUT)
            elif pb_prog == 'sander.APBS':
                pb_mdin = SanderAPBSInput(INPUT)
                pb_mdin2 = SanderAPBSInput(INPUT)
            else:
                pb_mdin = SanderPBSAInput(INPUT)
                pb_mdin2 = SanderPBSA2Input(INPUT)

            pb_mdin.write_input(pre + 'pb.mdin')
            pb_mdin2.write_input(pre + 'pb.mdin2')

    # end if decomprun

    if INPUT['entropy'] == 1:  # quasi-harmonic approximation input file
        trj_suffix = 'mdcrd'
        if INPUT['netcdf']: trj_suffix = 'nc'
        com_mask, rec_mask, lig_mask = prmtop_system.Mask('all', True)
        if not INPUT['mutant_only']:
            qh_in = QuasiHarmonicInput(com_mask, rec_mask, lig_mask,
                                       stability=stability, prefix=pre, trj_suffix=trj_suffix)
            qh_in.write_input(pre + 'cpptrajentropy.in')
        if INPUT['alarun']:
            qh_in = QuasiHarmonicInput(com_mask, rec_mask, lig_mask,
                                       stability=stability, prefix=pre + 'mutant_',
                                       trj_suffix=trj_suffix)
            qh_in.write_input(pre + 'mutant_cpptrajentropy.in')


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class NabInput(object):
    """
   Creates an input file for the Nab energy program. It must follow the format
   expected by the nab program:

   GB (or PB)
   variable1 = value1
   variable2 = value2
   ...
   variableN = valueN

   input_items is a dictionary of recognized keywords in the input file

   name_map is a dictionary that maps each key from input_items to the 
   corresponding key from the INPUT dictionary in gmx_MMPBSA
   """
    input_items = {'foo': 'bar'}  # This should be overridden in derived classes
    name_map = {'foo': 'orig'}  # This should be overridden in derived classes
    calculation_type = 'GB'  # This should be overridden in derived classes

    def __init__(self, INPUT):
        # Replace each of my input_items with the value held in INPUT if applicable
        # but if it's not in INPUT, don't pitch a fit (also ignore the possibility
        # that not every key is mapped in name_map, since only those that are in
        # INPUT will be mapped
      for key in list(self.input_items.keys()):
         try:
             self.input_items[key] = INPUT[self.name_map[key]]
         except KeyError:
             pass

    def write_input(self, filename):
        """ Writes the input file to filename """
        try:
            infile = open(filename, 'w')
        except TypeError:
            infile = filename

        infile.write('%s\n' % self.calculation_type)
        for key in list(self.input_items.keys()):
            infile.write('%s = %s\n' % (key, self.input_items[key]))
        infile.close()


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class GBNabInput(NabInput):
    """ The GB input file for the mmpbsa_py_energy NAB program """
    input_items = {'igb': 1,
                   'extdiel': 78.3,
                   'saltcon': 0.0,
                   'surften': 0.0072,
                   'rgbmax': 25.0}
    name_map = {'igb': 'igb',
                'extdiel': 'extdiel',
                'saltcon': 'saltcon',
                'surften': 'surften'}
    calculation_type = 'GB'


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class PBNabInput(NabInput):
    """ The PB input file for the mmpbsa_py_energy NAB program """
    input_items = {'inp': 2, 'smoothopt': 1, 'radiopt': 0,
                   'npbopt': 0, 'solvopt': 1, 'maxitn': 1000,
                   'nfocus': 2, 'bcopt': 5, 'eneopt': 2,
                   'fscale': 8, 'epsin': 1.0, 'epsout': 80.0,
                   'istrng': 0.0, 'dprob': 1.4, 'iprob': 2.0,
                   'accept': 0.001, 'fillratio': 4, 'space': 0.5,
                   'cutnb': 0, 'sprob': 0.557,
                   'cavity_surften': 0.0378, 'cavity_offset': -0.5692}
    name_map = {'inp': 'inp', 'radiopt': 'radiopt', 'maxitn': 'linit',
                'epsin': 'indi', 'epsout': 'exdi', 'istrng': 'istrng',
                'dprob': 'prbrad', 'fillratio': 'fillratio', 'space': 'scale',
                'cavity_surften': 'cavity_surften',
                'cavity_offset': 'cavity_offset'}
    calculation_type = 'PB'

    def __init__(self, INPUT):
        # Replace each of my input_items with the value held in INPUT if applicable
        # but if it's not in INPUT, don't pitch a fit (also ignore the possibility
        # that not every key is mapped in name_map, since only those that are in
        # INPUT will be mapped
        for key in list(self.input_items.keys()):
            try:
                self.input_items[key] = INPUT[self.name_map[key]]
            except KeyError:
                pass
        # Convert the ionic strength from M to mM for PBSA
        self.input_items['istrng'] = INPUT['istrng'] * 1000.0


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class SanderInput(object):
    """ Base class sander input file """
    program = 'sander'  # This runs with sander
    input_items = {'foo': 'bar'}  # replace this in derived classes
    name_map = {'foo': 'orig'}  # replace this in derived classes
    parent_namelist = {'foo': 'foo_namelist'}  # replace this in derived classes

    def __init__(self, INPUT):
        self.mdin = Mdin(self.program)
        self.mdin.title = 'File generated by gmx_MMPBSA'
        for key in list(self.input_items.keys()):
            # Skip ioutfm since it is handled explicitly later
            if key == 'ioutfm':
                continue
            try:
                self.mdin.change(self.parent_namelist[key], key,
                                 INPUT[self.name_map[key]])
            except KeyError:
                self.mdin.change(self.parent_namelist[key], key,
                                 self.input_items[key])
        self.mdin.change('cntrl', 'ioutfm', int(bool(INPUT['netcdf'])))

    def write_input(self, filename):
        """ Write the mdin file """
        self.mdin.write(filename)


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class SanderGBInput(SanderInput):
    """ GB sander input file """
    input_items = {'ntb': 0, 'cut': 999.0, 'nsnb': 99999,
                   'imin': 5, 'maxcyc': 1, 'ncyc': 0,
                   'igb': 5, 'saltcon': 0.0, 'intdiel': 1.0,
                   'gbsa': 0, 'extdiel': 80.0, 'surften': 0.0072,
                   'ioutfm': 0, 'idecomp': 0, 'offset': -999999.0,
                   'dec_verbose': 0, 'ifqnt': 0, 'qmmask': '',
                   'qm_theory': '', 'qmcharge': 0, 'qmgb': 2,
                   'qmcut': 999.0}
    parent_namelist = {'ntb': 'cntrl', 'cut': 'cntrl', 'nsnb': 'cntrl',
                       'imin': 'cntrl', 'maxcyc': 'cntrl', 'ncyc': 'cntrl',
                       'igb': 'cntrl', 'saltcon': 'cntrl', 'intdiel': 'cntrl',
                       'extdiel': 'cntrl', 'gbsa': 'cntrl', 'surften': 'cntrl',
                       'offset': 'cntrl', 'idecomp': 'cntrl', 'ioutfm': 'cntrl',
                       'dec_verbose': 'cntrl', 'ifqnt': 'cntrl', 'qmmask': 'qmmm',
                       'qm_theory': 'qmmm', 'qmcharge': 'qmmm', 'qmgb': 'qmmm',
                       'qmcut': 'qmmm'}
    name_map = {'igb': 'igb', 'saltcon': 'saltcon', 'intdiel': 'intdiel',
                'extdiel': 'extdiel', 'gbsa': 'gbsa', 'surften': 'surften',
                'idecomp': 'idecomp', 'ioutfm': 'netcdf', 'qmmask': 'qmmask',
                'qm_theory': 'qm_theory', 'ifqnt': 'ifqnt', 'qmcharge': 'qmcharge',
                'qmcut': 'qmcut', 'dec_verbose': 'dec_verbose'}


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class SanderPBSADECOMPInput(SanderInput):
    """ PB sander input file """
    input_items = {'ntb': 0, 'cut': 999.0, 'nsnb': 99999,
                   'imin': 5, 'maxcyc': 1, 'igb': 10,
                   'ipb': 2, 'inp': 2, 'idecomp': 0,
                   'epsin': 1.0, 'epsout': 80.0,
                   'istrng': 0.0, 'radiopt': 0, 'sprob': 0.557, 'dprob': 1.4,
                   'space': 0.5, 'maxitn': 1000, 'cavity_surften': 0.0378,
                   'cavity_offset': -0.5692, 'fillratio': 4.0,
                   'ioutfm': 0, 'dec_verbose': 0}
    name_map = {'inp': 'inp', 'idecomp': 'idecomp', 'epsin': 'indi',
                'epsout': 'exdi', 'istrng': 'istrng', 'radiopt': 'radiopt',
                'dprob': 'prbrad', 'space': 'scale', 'maxitn': 'linit',
                'cavity_surften': 'cavity_surften',
                'cavity_offset': 'cavity_offset', 'fillratio': 'fillratio',
                'ioutfm': 'netcdf', 'dec_verbose': 'dec_verbose'}
    parent_namelist = {'ntb': 'cntrl', 'cut': 'cntrl', 'nsnb': 'cntrl',
                       'imin': 'cntrl', 'maxcyc': 'cntrl', 'igb': 'cntrl',
                       'ipb': 'cntrl', 'inp': 'cntrl', 'idecomp': 'cntrl',
                       'epsin': 'pb', 'epsout': 'pb',
                       'istrng': 'pb', 'radiopt': 'pb', 'sprob': 'pb', 'dprob': 'pb',
                       'space': 'pb', 'maxitn': 'pb', 'cavity_offset': 'pb',
                       'cavity_surften': 'pb', 'fillratio': 'pb',
                       'ioutfm': 'cntrl', 'dec_verbose': 'cntrl'}

    def __init__(self, INPUT):
        # We need to change istrng to mM (from M).
        SanderInput.__init__(self, INPUT)
        self.mdin.change('pb', 'istrng', INPUT['istrng'] * 1000)


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class SanderPBSAInput(SanderInput):
    """ PB sander input file """
    input_items = {'ntb': 0, 'cut': 999.0, 'nsnb': 99999,
                   'imin': 5, 'maxcyc': 1, 'igb': 10,
                   'ipb': 2, 'inp': 2, 'idecomp': 0,
                   'epsin': 1.0, 'epsout': 80.0,
                   'istrng': 0.0, 'radiopt': 0, 'sprob': 0.557, 'dprob': 1.4,
                   'space': 0.5, 'maxitn': 1000, 'cavity_surften': 0.0378,
                   'cavity_offset': -0.5692, 'fillratio': 4.0,
                   'ioutfm': 0, 'dec_verbose': 0,
                   'ntx': 1, 'epsmem': 1.0, 'membraneopt': 0, 'sasopt': 0,
                   'mthick': 40, 'maxarcdot': 1500, 'solvopt': 2, 'nfocus': 2, 'bcopt': 5,
                   'eneopt': 2, 'frcopt': 0, 'cutfd': 5.0, 'cutnb': 0.0,
                   'mctrdz': 0.0, 'poretype': 1, 'npbverb': 0}
    name_map = {'inp': 'inp', 'idecomp': 'idecomp', 'epsin': 'indi',
                'epsout': 'exdi', 'istrng': 'istrng', 'radiopt': 'radiopt',
                'dprob': 'prbrad', 'space': 'scale', 'maxitn': 'linit',
                'cavity_surften': 'cavity_surften',
                'cavity_offset': 'cavity_offset', 'fillratio': 'fillratio',
                'ioutfm': 'netcdf', 'dec_verbose': 'dec_verbose',
                'ipb': 'ipb', 'epsmem': 'emem', 'membraneopt': 'memopt',
                'sasopt': 'sasopt', 'mthick': 'mthick', 'maxarcdot': 'maxarcdot', 'solvopt': 'solvopt',
                'nfocus': 'nfocus', 'bcopt': 'bcopt', 'eneopt': 'eneopt',
                'frcopt': 'frcopt', 'cutfd': 'cutfd', 'cutnb': 'cutnb',
                'mctrdz': 'mctrdz', 'poretype': 'poretype', 'npbverb': 'npbverb'}
    parent_namelist = {'ntb': 'cntrl', 'cut': 'cntrl', 'nsnb': 'cntrl',
                       'imin': 'cntrl', 'maxcyc': 'cntrl', 'igb': 'cntrl',
                       'ipb': 'cntrl', 'inp': 'cntrl', 'idecomp': 'cntrl',
                       'epsin': 'pb', 'epsout': 'pb',
                       'istrng': 'pb', 'radiopt': 'pb', 'sprob': 'pb', 'dprob': 'pb',
                       'space': 'pb', 'maxitn': 'pb', 'cavity_offset': 'pb',
                       'cavity_surften': 'pb', 'fillratio': 'pb',
                       'ioutfm': 'cntrl', 'dec_verbose': 'cntrl',
                       'ntx': 'cntrl', 'epsmem': 'pb', 'membraneopt': 'pb',
                       'sasopt': 'pb', 'mthick': 'pb', 'maxarcdot': 'pb', 'solvopt': 'pb',
                       'nfocus': 'pb', 'bcopt': 'pb', 'eneopt': 'pb', 'frcopt': 'pb',
                       'cutfd': 'pb', 'cutnb': 'pb', 'mctrdz': 'pb',
                       'poretype': 'pb', 'npbverb': 'pb'}

    def __init__(self, INPUT):
        # We need to change istrng to mM (from M).
        SanderInput.__init__(self, INPUT)
        self.mdin.change('pb', 'istrng', INPUT['istrng'] * 1000)


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class SanderPBSA2Input(SanderInput):
    """ PB sander input file """
    input_items = {'ntb': 0, 'cut': 999.0, 'nsnb': 99999,
                   'imin': 5, 'maxcyc': 1, 'igb': 10,
                   'ipb': 2, 'inp': 2, 'idecomp': 0,
                   'epsin': 1.0, 'epsout': 80.0,
                   'istrng': 0.0, 'radiopt': 0, 'sprob': 0.557, 'dprob': 1.4,
                   'space': 0.5, 'maxitn': 1000, 'cavity_surften': 0.0378,
                   'cavity_offset': -0.5692, 'fillratio': 4.0,
                   'ioutfm': 0, 'dec_verbose': 0,
                   'ntx': 1, 'epsmem': 1.0, 'membraneopt': 0, 'sasopt': 0,
                   'mthick': 40, 'maxarcdot': 1500, 'solvopt': 2, 'nfocus': 2, 'bcopt': 5,
                   'eneopt': 2, 'frcopt': 0, 'cutfd': 5.0, 'cutnb': 0.0,
                   'mctrdz': 0.0, 'poretype': 1, 'npbverb': 0}
    name_map = {'inp': 'inp', 'idecomp': 'idecomp', 'epsin': 'indi',
                'epsout': 'exdi', 'istrng': 'istrng', 'radiopt': 'radiopt',
                'dprob': 'prbrad', 'space': 'scale', 'maxitn': 'linit',
                'cavity_surften': 'cavity_surften',
                'cavity_offset': 'cavity_offset', 'fillratio': 'fillratio',
                'ioutfm': 'netcdf', 'dec_verbose': 'dec_verbose',
                'ipb': 'ipb', 'epsmem': 'emem', 'membraneopt': 'memoptzero',
                'sasopt': 'sasopt', 'mthick': 'mthick', 'maxarcdot': 'maxarcdot', 'solvopt': 'solvopt',
                'nfocus': 'nfocus', 'bcopt': 'bcopt', 'eneopt': 'eneopt',
                'frcopt': 'frcopt', 'cutfd': 'cutfd', 'cutnb': 'cutnb',
                'mctrdz': 'mctrdz', 'poretype': 'poretype', 'npbverb': 'npbverb'}
    parent_namelist = {'ntb': 'cntrl', 'cut': 'cntrl', 'nsnb': 'cntrl',
                       'imin': 'cntrl', 'maxcyc': 'cntrl', 'igb': 'cntrl',
                       'ipb': 'cntrl', 'inp': 'cntrl', 'idecomp': 'cntrl',
                       'epsin': 'pb', 'epsout': 'pb',
                       'istrng': 'pb', 'radiopt': 'pb', 'sprob': 'pb', 'dprob': 'pb',
                       'space': 'pb', 'maxitn': 'pb', 'cavity_offset': 'pb',
                       'cavity_surften': 'pb', 'fillratio': 'pb',
                       'ioutfm': 'cntrl', 'dec_verbose': 'cntrl',
                       'ntx': 'cntrl', 'epsmem': 'pb', 'membraneopt': 'pb',
                       'sasopt': 'pb', 'mthick': 'pb', 'maxarcdot': 'pb', 'solvopt': 'pb',
                       'nfocus': 'pb', 'bcopt': 'pb', 'eneopt': 'pb', 'frcopt': 'pb',
                       'cutfd': 'pb', 'cutnb': 'pb', 'mctrdz': 'pb',
                       'poretype': 'pb', 'npbverb': 'pb'}

    def __init__(self, INPUT):
        # We need to change istrng to mM (from M).
        SanderInput.__init__(self, INPUT)
        self.mdin.change('pb', 'istrng', INPUT['istrng'] * 1000)


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class SanderAPBSInput(SanderInput):
    """ PB sander input file using APBS as the PB solver """
    program = 'sander.APBS'
    input_items = {'ntb': 0, 'cut': 999.0, 'nsnb': 99999,
                   'imin': 5, 'maxcyc': 1, 'igb': 6,
                   'apbs_print': 1, 'calc_type': 0, 'cmeth': 1,
                   'bcfl': 2, 'srfm': 2, 'chgm': 1, 'pdie': 1.0,
                   'sdie': 80.0, 'srad': 1.4, 'nion': 2, 'ionq': '1.0,-1.0',
                   'ionc': '0.0,0.0', 'ionrr': '2.0,2.0', 'radiopt': 0,
                   'calcforce': 0, 'calcnpenergy': 1, 'grid': '0.5,0.5,0.5',
                   'gamma': 0.00542, 'ioutfm': 0}
    name_map = {'inp': 'inp', 'idecomp': 'idecomp', 'pdie': 'indi',
                'sdie': 'exdi', 'ionc': 'istrng', 'radiopt': 'radiopt',
                'srad': 'prbrad', 'grid': 'scale', 'gamma': 'cavity_surften',
                'ioutfm': 'netcdf'}
    parent_namelist = {'ntb': 'cntrl', 'cut': 'cntrl', 'nsnb': 'cntrl',
                       'imin': 'cntrl', 'maxcyc': 'cntrl', 'igb': 'cntrl',
                       'ipb': 'cntrl', 'inp': 'cntrl', 'idecomp': 'cntrl',
                       'ioutfm': 'cntrl', 'apbs_print': 'apbs', 'calc_type': 'apbs',
                       'cmeth': 'apbs', 'bcfl': 'apbs', 'srfm': 'apbs',
                       'chgm': 'apbs', 'pdie': 'apbs', 'sdie': 'apbs',
                       'srad': 'apbs', 'nion': 'apbs', 'ionq': 'apbs',
                       'ionc': 'apbs', 'ionrr': 'apbs', 'radiopt': 'apbs',
                       'calcforce': 'apbs', 'grid': 'apbs', 'gamma': 'apbs',
                       'calcnpenergy': 'apbs'}

    def __init__(self, INPUT):
        SanderInput.__init__(self, INPUT)
        # We also have to make some modifications specific to sander.APBS
        self.mdin.change('apbs', 'grid', '%s,%s,%s' % (INPUT['scale'],
                                                       INPUT['scale'], INPUT['scale']))
        self.mdin.change('apbs', 'ionc', '%s,%s' % (INPUT['istrng'],
                                                    INPUT['istrng']))
        self.mdin.change('apbs', 'gamma', INPUT['cavity_surften'] * 4.184)


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class SanderGBDecomp(SanderGBInput):
    """ GB decomposition input file for sander. In addition to the INPUT dictionary,
       this class also needs several GROUP cards to define the 
   """

    def __init__(self, INPUT, *cards):
        SanderGBInput.__init__(self, INPUT)
        for card in cards:
            self.mdin.AddCard(card[0], card[1])


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class SanderPBDecomp(SanderPBSADECOMPInput):
    """ PB decomposition input file for sander """

    def __init__(self, INPUT, *cards):
        SanderPBSADECOMPInput.__init__(self, INPUT)
        for card in cards:
            self.mdin.AddCard(card[0], card[1])


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class QuasiHarmonicInput(object):
    """ Write a cpptraj input file to do a quasi-harmonic entropy calculation """

    def __init__(self, com_mask, rec_mask, lig_mask, stability=False,
                 prefix='_MMPBSA_', trj_suffix='mdcrd'):
        self.file_string = ("trajin %scomplex.%s\n" % (prefix, trj_suffix) +
                            "reference %savgcomplex.pdb\n" % (prefix) +
                            "rms mass reference %s\n" % (com_mask) +
                            "matrix mwcovar name comp.matrix %s\n" % (com_mask) +
                            "analyze matrix comp.matrix out " +
                            "%scomplex_entropy.out thermo reduce\n" % (prefix))
        if not stability:
            self.file_string = (self.file_string +
                                "rms mass reference %s\n" % (rec_mask) +
                                "matrix mwcovar name rec.matrix %s\n" % (rec_mask) +
                                "analyze matrix rec.matrix out " +
                                "%sreceptor_entropy.out thermo reduce\n" % (prefix) +
                                "rms mass reference %s\n" % (lig_mask) +
                                "matrix mwcovar name lig.matrix %s\n" % (lig_mask) +
                                "analyze matrix lig.matrix out " +
                                "%sligand_entropy.out thermo reduce\n" % (prefix))

    def write_input(self, filename):
        """ Writes the input file """
        open(filename, 'w').write(self.file_string)

# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+



# ##############################################################################
#                           GPLv3 LICENSE INFO                                 #
#                                                                              #
#  Copyright (C) 2020  Mario S. Valdes-Tresanco and Mario E. Valdes-Tresanco   #
#  Copyright (C) 2014  Jason Swails, Bill Miller III, and Dwight McGee         #
#                                                                              #
#   Project: https://github.com/Valdes-Tresanco-MS/gmx_MMPBSA                  #
#                                                                              #
#   This program is free software; you can redistribute it and/or modify it    #
#  under the terms of the GNU General Public License version 3 as published    #
#  by the Free Software Foundation.                                            #
#                                                                              #
#  This program is distributed in the hope that it will be useful, but         #
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    #
#  for more details.                                                           #
# ##############################################################################

from setuptools import setup
import versioneer

with open("README.md", "r") as f:
    LONG_DESCRIPTION = f.read()

def get_version(rel_path):
    with open(rel_path) as vfile:
        for line in vfile.readlines():
            if line.startswith('__version__'):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
        else:
            raise RuntimeError("Unable to find version string.")

setup(
    name='gmx_MMPBSA',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    packages=['GMXMMPBSA'],
    package_data={"GMXMMPBSA": ["data/*", 'GMXMMPBSA.sh']},
    license='GPLv3',
    author='Mario S. Valdes-Trasanco and Mario E. Valdes-Tresanco ',
    author_email='mariosergiovaldes145@gmail.com',
    maintainer='Mario S. Valdes-Trasanco',
    maintainer_email='mariosergiovaldes145@gmail.com',
    url='https://github.com/Valdes-Tresanco-MS/gmx_MMPBSA',
    description='gmx_MMPBSA is a new tool aid to perform end-state free energy'
                'calculations with GROMACS files.',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    keywords=['gmx_MMPBSA', 'MMPBSA', 'MMGBSA', 'GROMACS', 'AmberTools'],
    entry_points={
        "console_scripts": [
            "gmx_MMPBSA=GMXMMPBSA.app:gmxmmpbsa",
            "gmx_MMPBSA_ana=GMXMMPBSA.app:gmxmmpbsa_ana"]}
)



"""
 This package contains all of the functions for gmx_MMPBSA that it
 needs to run smoothly.
"""

# ##############################################################################
#                           GPLv3 LICENSE INFO                                 #
#                                                                              #
#  Copyright (C) 2020  Mario S. Valdes-Tresanco and Mario E. Valdes-Tresanco   #
#  Copyright (C) 2014  Jason Swails, Bill Miller III, and Dwight McGee         #
#                                                                              #
#   Project: https://github.com/Valdes-Tresanco-MS/gmx_MMPBSA                  #
#                                                                              #
#   This program is free software; you can redistribute it and/or modify it    #
#  under the terms of the GNU General Public License version 3 as published    #
#  by the Free Software Foundation.                                            #
#                                                                              #
#  This program is distributed in the hope that it will be useful, but         #
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    #
#  for more details.                                                           #
# ##############################################################################

__all__ = ['alamdcrd', 'amber_outputs', 'analyzer', 'API', 'app', 'calculation', 'commandlineparser', 'createinput',
           'exceptions', 'fake_mpi', 'findprogs', 'infofile', 'input_parser', 'main', 'make_top', 'make_trajs',
           'output_file', 'parm_setup', 'timer', 'utils']

__author__ = "Mario S. Valdes Tresanco and Mario E. Valdes Tresanco"
__license__ = "GPLv3"
__mmpbsa_author__ = "Jason Swails, Dwight McGee, and Bill Miller III"
__mmpbsa_version__ = "16.0"
__ambertools_version__ = "20"

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



"""
This module is responsible for finding the necessary programs for all of the
desired calculations in the gmx_MMPBSA calculation.

Methods:
         find_progs(INPUT): Determines which programs are needed and sees if
                            they can be found. Returns dictionary of programs
         which(program): Internal; searches AMBERHOME/bin and optionally
                         PATH to see if a program can be found

Classes: ExternProg: External program class
"""

# ##############################################################################
#                           GPLv3 LICENSE INFO                                 #
#                                                                              #
#  Copyright (C) 2020  Mario S. Valdes-Tresanco and Mario E. Valdes-Tresanco   #
#  Copyright (C) 2014  Jason Swails, Bill Miller III, and Dwight McGee         #
#                                                                              #
#   Project: https://github.com/Valdes-Tresanco-MS/gmx_MMPBSA                  #
#                                                                              #
#   This program is free software; you can redistribute it and/or modify it    #
#  under the terms of the GNU General Public License version 3 as published    #
#  by the Free Software Foundation.                                            #
#                                                                              #
#  This program is distributed in the hope that it will be useful, but         #
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    #
#  for more details.                                                           #
# ##############################################################################

from GMXMMPBSA.exceptions import GMXMMPBSA_ERROR
import logging
#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

def find_progs(INPUT):
    """ Find the necessary programs based in the user INPUT """
    # List all of the used programs with the conditions that they are needed
    logging.info('Checking external programs...')

    used_progs = { 'cpptraj' : True,

                   'tleap': True, 'parmchk2': True,
                   'mmpbsa_py_energy' : ((INPUT['pbrun'] or INPUT['gbrun']) and not
                                         (INPUT['use_sander'] or INPUT['decomprun'])),
                   'sander' : (INPUT['decomprun'] or INPUT['use_sander'] or INPUT['ifqnt'] == 1),
                   'sander.APBS' : INPUT['sander_apbs'] == 1,
                   'mmpbsa_py_nabnmode' : INPUT['nmoderun'],
                   'rism3d.snglpnt' : INPUT['rismrun']
                   }
    gro_exe = {
        'gmx5': [
            # look for any available gromacs executable
            'gmx', 'gmx_mpi', 'gmx_d', 'gmx_mpi_d'],
        'gmx4': [
            # look for gromacs 4.x
            'make_ndx', 'trjconv', 'editconf']}

    # The returned dictionary:
    my_progs = {}

    search_path = True
    force_path = INPUT['gmx_path']

    for prog in list(used_progs.keys()):
        my_progs[prog] = ExternProg(prog, used_progs[prog], search_path).full_path
        if used_progs[prog]:
            if not my_progs[prog]:
                GMXMMPBSA_ERROR('Could not find necessary program [%s]' % prog)
            logging.info('%s found! Using %s' % (prog, str(my_progs[prog])))

    if force_path:
        search_path = False
    g5 = False
    for v in gro_exe:
        if v == 'gmx5':
            for prog in gro_exe[v]:
                exe = ExternProg(prog, True, search_path, force_path)
                if exe.full_path:
                    logging.info('Using GROMACS version > 5.x.x!')
                    my_progs['make_ndx'] = [exe.full_path, 'make_ndx']
                    my_progs['editconf'] = [exe.full_path, 'editconf']
                    my_progs['trjconv'] = [exe.full_path, 'trjconv']
                    g5 = True
                    logging.info('%s found! Using %s' % (prog, exe.full_path))
                    break
            if g5:
                break
        else:
            c = 0
            logging.info('Using GROMACS version 4.x.x!')
            for prog in gro_exe[v]:
                exe = ExternProg(prog, True, search_path, force_path)
                if exe.full_path:
                    c += 1
                    my_progs[prog] = [exe.full_path]
                    logging.info('%s found! Using %s' % (prog, str(my_progs[prog])))

    if 'make_ndx' not in my_progs or 'editconf' not in my_progs or 'trjconv' not in my_progs:
        GMXMMPBSA_ERROR('Could not find necessary program [ GROMACS ]')
    logging.info('Checking external programs...Done.\n')
    return my_progs

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

def which(program, search_path=False, force_path=None):
    """ Searches for a program in $AMBERHOME first, then PATH if we allow
        PATH searching.
    """
    import os

    def is_exe(filename):
        return os.path.exists(filename) and os.access(filename, os.X_OK)

    def get_amberhome():
        ambhome = os.getenv('AMBERHOME')
        if ambhome == None:
            GMXMMPBSA_ERROR('AMBERHOME is not set!')
        return ambhome

    # Check to see that a path was provided in the program name
    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program): return program

    # If not (and it generally isn't), then look in AMBERHOME
    amberhome = get_amberhome()

    if is_exe(os.path.join(amberhome, 'bin', program)):
        return os.path.join(amberhome, 'bin', program)

    if force_path:
        exe_file = os.path.join(force_path, program)
        if is_exe(exe_file):
            return exe_file  # if it's executable, return the file

    # If we can search the path, look through it
    if search_path:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file # if it's executable, return the file

    return None  # if program can still not be found... return None

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class ExternProg(object):
    """ An external program """
    def __init__(self, prog_name, needed=False, search_path=False, force_path=None):
        # We need to initialize the program name that we're looking for, whether
        # or not it is necessary by default, and the conditions upon which we
        # must look for it. The full_path is the attribute that stores this
        # program's full PATH name. We also need to know if we are going to search
        # through the whole PATH or just AMBERHOME.
        self.prog_name = prog_name
        self.full_path = None
        self.search_path = search_path
        self.force_path = force_path
        if needed: self.find()

    def find(self):
        """ Find the program """
        self.full_path = which(self.prog_name, self.search_path, self.force_path)

    def __str__(self):
        if self.full_path:
            return self.full_path
        else:
            return self.prog_name

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+



# -*- coding: utf-8 -*-
"""
This module contains classes and such that are responsible for parsing
command-line arguments for gmx_MMPBSA.  All of the files specified for use
in gmx_MMPBSA will be assigned as attributes to the returned class.

"""

# ##############################################################################
#                           GPLv3 LICENSE INFO                                 #
#                                                                              #
#  Copyright (C) 2020  Mario S. Valdes-Tresanco and Mario E. Valdes-Tresanco   #
#  Copyright (C) 2014  Jason Swails, Bill Miller III, and Dwight McGee         #
#                                                                              #
#   Project: https://github.com/Valdes-Tresanco-MS/gmx_MMPBSA                  #
#                                                                              #
#   This program is free software; you can redistribute it and/or modify it    #
#  under the terms of the GNU General Public License version 3 as published    #
#  by the Free Software Foundation.                                            #
#                                                                              #
#  This program is distributed in the hope that it will be useful, but         #
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    #
#  for more details.                                                           #
# ##############################################################################

import os
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from pathlib import Path
from GMXMMPBSA import __version__, __mmpbsa_version__, __ambertools_version__
from GMXMMPBSA.exceptions import GMXMMPBSA_ERROR


class OptionList(object):
    """
    Just a container to hold the command-line options. Necessary when reading in
    a gmx_MMPBSA info file to have a container to load the results from the
    parser.
    """
    pass


if os.getenv('AMBERHOME'):
    rism = os.path.join(os.getenv('AMBERHOME'), 'dat', 'mmpbsa', 'spc.xvv')
else:
    rism = None

def check_arg(str_suffix, path=False):
    def decorator(f):
        def gmx_MMPBSA_file(*args):
            result = f(*args)
            if path and not Path(result).exists():
                GMXMMPBSA_ERROR(f'{result} do not exist or is inaccessible')
            if not Path(result).suffix in str_suffix:
                GMXMMPBSA_ERROR(f'{result} does not correspond to the required structure format {str_suffix}')
            return result

        return gmx_MMPBSA_file

    return decorator


@check_arg(['.tpr', '.pdb', '.gro'], True)
def structure(arg):
    return arg


@check_arg(['.pdb'], True)
def pdb(arg):
    return arg


@check_arg(['.xtc', '.trr', '.pdb'], True)
def trajectory(arg):
    return arg


@check_arg(['.top'], True)
def topology(arg):
    return arg


@check_arg(['.mol2'], True)
def mol2(arg):
    return arg


@check_arg(['.ndx'], True)
def index(arg):
    return arg


description = '''gmx_MMPBSA is an effort to implement the GB/PB and others calculations in GROMACS. This program is 
                 an adaptation of Amber's MMPBSA.py and essentially works as such. As gmx_MMPBSA adapts MMPBSA.py, 
                 since it has all the resources of this script and work with any GROMACS version.'''

complex_group_des = '''Complex files and info that are needed to perform the calculation. If the receptor and / or the
                        ligand info is not defined, we generate them from that of the complex.'''

receptor_group_des = '''Receptor files and info that are needed to perform the calculation. If the receptor info is not 
                         defined, we generate it from that of the complex.'''

ligand_group_des = '''Ligand files and info that are needed to perform the calculation. If the ligand are not defined, 
                       we generate it from that of the complex.'''
# Set up the MM/PBSA parser here. It clutters up the MMPBSA_App to do it there

# noinspection PyTypeChecker
parser = ArgumentParser(epilog=f'''This program will calculate binding free energies using end-state free energy 
                                    methods on an ensemble of snapshots using a variety of implicit solvent 
                                    models.\nBased on MMPBSA.py (version {__mmpbsa_version__}) and 
                                    AmberTools{__ambertools_version__}''',
                        description=description, formatter_class=ArgumentDefaultsHelpFormatter)
parser.add_argument('-v', '--version', action='version',
                    version='''%%(prog)s %s based on MMPBSA version %s and AmberTools %s''' %
                            (__version__, __mmpbsa_version__, __ambertools_version__))
parser.add_argument('--input-file-help', dest='infilehelp', action='store_true',
                    help='Print all available options in the input file.',
                    default=False)
group = parser.add_argument_group('Miscellaneous Options')
group.add_argument('-O', '--overwrite', default=False, action='store_true',
                   help='Allow output files to be overwritten', dest='overwrite')
group.add_argument('-prefix', dest='prefix', default='_GMXMMPBSA_',
                   metavar='<file prefix>',
                   help='Prefix for intermediate files.')
group = parser.add_argument_group('Input and Output Files', '''These options specify the input files and optional 
output files.''')
group.add_argument('-i', dest='input_file', metavar='FILE', help='MM/PBSA input file.')
group.add_argument('-xvvfile', dest='xvvfile', help='XVV file for 3D-RISM.', default=rism)
group.add_argument('-o', dest='output_file', default='FINAL_RESULTS_MMPBSA.dat', metavar='FILE',
                   help='Output file with MM/PBSA statistics.')
group.add_argument('-do', dest='decompout', metavar='FILE', default='FINAL_DECOMP_MMPBSA.dat',
                   help='Output file for decomposition statistics summary.')
group.add_argument('-eo', dest='energyout', metavar='FILE',
                   help='''CSV-format output of all energy terms for every frame
                  in every calculation. File name forced to end in [.csv].
                  This file is only written when specified on the
                  command-line.''')
group.add_argument('-deo', dest='dec_energies', metavar='FILE',
                   help='''CSV-format output of all energy terms for each printed
                  residue in decomposition calculations. File name forced to end
                  in [.csv]. This file is only written when specified on the
                  command-line.''')
group.add_argument('-nogui', dest='gui', action='store_false', default=True,
                   help='Open charts application (gmx_MMPBSA_ana) when all calculations finished')
group.add_argument('-s', '--stability', dest='stability', action='store_true', default=False,
                   help='''Perform stability calculation. Only the complex parameters are required. If
                         ligand is non-Protein (small molecule) type, then ligand *.mol2 file is 
                         required. In any other case receptor and ligand parameters will be ignored.
                         See description bellow''')

group = parser.add_argument_group('Complex', complex_group_des)
group.add_argument('-cs', dest='complex_tpr', metavar='<Structure File>', default=None, type=structure,
                   help='''Structure file of the complex. If it is Protein-Ligand (small molecule)
                         complex, make sure that you define -lm option. See -lm description below
                         Allowed formats: *.tpr (recommended), *.pdb, *.gro''')
group.add_argument('-ci', dest='complex_index', metavar='<Index File>', default=None, type=index,
                   help='Index file of the bound complex.')
group.add_argument('-cg', dest='complex_groups', metavar='index', nargs=2, default=None, type=int,
                   help='Groups of receptor and ligand in complex index file. The notation is as follows: "-cg '
                        '<Receptor group> <Ligand group>", ie. -cg 1 13')
group.add_argument('-ct', dest='complex_trajs', nargs='*', metavar='TRJ', type=trajectory,
                   help='''Input trajectories of the complex. Make sure the trajectory is fitted and
                         pbc have been removed. Allowed formats: *.xtc (recommended), *.trr, *.pdb
                         (specify as many as you'd like).''')
group.add_argument('-cp', dest='complex_top', metavar='<Topology>', default=None, type=topology,
                   help='''Topology file of the complex.''')
group.add_argument('-cr', dest='reference_structure', metavar='<PDB File>', default=None, type=pdb,
                   help='''Complex Reference Structure file. This option is optional but recommended 
                         (Use the PDB file used to generate the topology in GROMACS). If not defined,
                         the chains ID assignment (if the structure used in -cs does not have chain
                         IDs) will be done automatically according to the structure (can generate
                         inconsistencies).''')

group = parser.add_argument_group('Receptor', receptor_group_des)
group.add_argument('-rs', dest='receptor_tpr', metavar='<Structure File>', default=None, type=structure,
                   help='''Structure file of the unbound receptor for multiple trajectory approach.
                         Allowed formats: *.tpr (recommended), *.pdb, *.gro''')
group.add_argument('-ri', dest='receptor_index', metavar='<Index File>', default=None, type=index,
                   help='Index file of the unbound receptor.')
group.add_argument('-rg', dest='receptor_group', metavar='index', default=None, type=int,
                   help='''Receptor group in receptor index file. Notation: "-lg <Receptor group>", 
                         e.g. -rg 1''')
group.add_argument('-rt', dest='receptor_trajs', nargs='*', metavar='TRJ', type=trajectory,
                   help='''Input trajectories of the unbound receptor for multiple trajectory approach.
                         Allowed formats: *.xtc (recommended), *.trr, *.pdb, *.gro (specify as many
                         as you'd like).''')
group.add_argument('-rp', dest='receptor_top', metavar='<Topology>', default=None, type=topology,
                   help='''Topology file of the receptor.''')

group = parser.add_argument_group('Ligand', ligand_group_des)
group.add_argument('-lm', dest='ligand_mol2', metavar='<Structure File>', default=None, type=mol2,
                   help='''A *.mol2 file of the unbound ligand used to parametrize ligand for GROMACS
                         using Anetchamber. Must be defined if Protein-Ligand (small molecule) 
                         complex was define. No needed for Proteins, DNA, RNA, Ions and Glycans.
                         Antechamber output *.mol2 is recommended.''')
group.add_argument('-ls', dest='ligand_tpr', metavar='<Structure File>', default=None, type=structure,
                   help='''Structure file of the unbound ligand. If ligand is a small molecule, make 
                         sure that you define above -lm option. Allowed formats: *.tpr (recommended),
                         *.pdb, *.gro''')
group.add_argument('-li', dest='ligand_index', metavar='<Index File>', type=index,
                   default=None, help='Index file of the unbound ligand. Only if tpr file was define in -ls.')
group.add_argument('-lg', dest='ligand_group', metavar='index', default=None, type=int,
                   help='''Ligand group in ligand index file. Notation: "-lg <Ligand group>", 
                         e.g. -lg 13''')
group.add_argument('-lt', dest='ligand_trajs', nargs='*', metavar='TRJ', type=trajectory,
                   help='''Input trajectories of the unbound ligand for multiple trajectory approach. 
                         Allowed formats: *.xtc (recommended), *.trr, *.pdb, *.gro (specify as many
                         as you'd like).''')
group.add_argument('-lp', dest='ligand_top', metavar='<Topology>', default=None, type=topology,
                   help='''Topology file of the ligand.''')

group = parser.add_argument_group('Miscellaneous Actions')
group.add_argument('-make-mdins', dest='make_mdins', default=False,
                   action='store_true', help='''Create the input files for each
                  calculation and quit. This allows you to modify them and
                  re-run using -use-mdins''')
group.add_argument('-use-mdins', dest='use_mdins', default=False,
                   action='store_true', help='''Use existing input files for each
                  calculation. If they do not exist with the appropriate names,
                  %(prog)s will quit in error.''')
group.add_argument('-rewrite-output', dest='rewrite_output', default=False,
                   action='store_true', help='''Do not re-run any calculations,
                  just parse the output files from the previous calculation and
                  rewrite the output files.''')
group.add_argument('--clean', dest='clean', action='store_true', default=False,
                   help='''Clean temporary files and quit.''')

#### GUI parser
anaparser = ArgumentParser(epilog='''This program is part of gmx_MMPBSA and will show a workspace with 
                            charts to analyze the results''',
                           description=description,
                           formatter_class=ArgumentDefaultsHelpFormatter)
anaparser.add_argument('-v', '--version', action='version',
                       version='%%(prog)s %s based on MMPBSA version %s' % (__version__, __mmpbsa_version__))
group = anaparser.add_argument_group('Info file')
group.add_argument('-p', '--path', dest='path', help='Path to gmx_MMPBSA info file', required=True,
                   default=None)



# ##############################################################################
#                           GPLv3 LICENSE INFO                                 #
#                                                                              #
#  Copyright (C) 2020  Mario S. Valdes-Tresanco and Mario E. Valdes-Tresanco   #
#  Copyright (C) 2014  Jason Swails, Bill Miller III, and Dwight McGee         #
#                                                                              #
#   Project: https://github.com/Valdes-Tresanco-MS/gmx_MMPBSA                  #
#                                                                              #
#   This program is free software; you can redistribute it and/or modify it    #
#  under the terms of the GNU General Public License version 3 as published    #
#  by the Free Software Foundation.                                            #
#                                                                              #
#  This program is distributed in the hope that it will be useful, but         #
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    #
#  for more details.                                                           #
# ##############################################################################
import sys
import os
from pathlib import Path
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from PyQt5.QtGui import *
from GMXMMPBSA import API
from GMXMMPBSA.make_trajs import Trajectory
from GMXMMPBSA.utils import PDB
from GMXMMPBSA.findprogs import find_progs
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT
import matplotlib.backend_bases
from matplotlib.backends import qt_compat
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
import numpy as np
from math import ceil

plt.style.use('seaborn')

std_aa = ['ALA', 'ARG', 'ASN', 'ASP', 'ASH', 'CYS', 'CYX', 'GLN', 'GLU', 'GLH', 'GLY', 'HID', 'HIE', 'HIP', 'ILE',
          'LEU', 'LYS', 'LYN', 'MET', 'PHE', 'PRO', 'SER', 'THR', 'TRP', 'TYR', 'VAL', 'HIS']


class NavigationToolbar(NavigationToolbar2QT):
    """
    overwrite NavigatorToolbar to get control over save action.
    Now we can set custom dpi value
    """

    def _init_toolbar(self):
        pass

    def save_figure(self, *args):
        filetypes = self.canvas.get_supported_filetypes_grouped()
        sorted_filetypes = sorted(filetypes.items())
        default_filetype = self.canvas.get_default_filetype()

        startpath = os.path.expanduser(
            matplotlib.rcParams['savefig.directory'])
        start = os.path.join(startpath, self.canvas.get_default_filename())
        filters = []
        selectedFilter = None
        for name, exts in sorted_filetypes:
            exts_list = " ".join(['*.%s' % ext for ext in exts])
            filter = '%s (%s)' % (name, exts_list)
            if default_filetype in exts:
                selectedFilter = filter
            filters.append(filter)
        filters = ';;'.join(filters)

        fname, filter = qt_compat._getSaveFileName(
            self.canvas.parent(), "Choose a filename to save to", start,
            filters, selectedFilter)
        if fname:
            # Save dir for next time, unless empty str (i.e., use cwd).
            if startpath != "":
                matplotlib.rcParams['savefig.directory'] = (
                    os.path.dirname(fname))
            try:
                self.canvas.figure.savefig(fname, dpi=300)
            except Exception as e:
                QMessageBox.critical(
                    self, "Error saving file", str(e),
                    QMessageBox.Ok, QMessageBox.NoButton)


class Charts(QMdiSubWindow):
    def __init__(self, *args):
        super(Charts, self).__init__(*args)
        self.item = None
        self.col = None

    def closeEvent(self, closeEvent: QCloseEvent) -> None:
        self.item.setCheckState(self.col, Qt.Unchecked)


class CustomItem(QTreeWidgetItem):
    def __init__(self, *args):
        super(CustomItem, self).__init__(*args)
        self.subwindows = {1: None, 2: None}
        self.dataperframe = None
        self.datamean = None
        self.name = None
        self.keyslist = []


class MplCanvas(FigureCanvasQTAgg):

    def __init__(self, parent=None, width=5, height=4, dpi=100):
        fig = Figure(figsize=(width, height), dpi=dpi)
        self.axes = fig.add_subplot(111)
        super(MplCanvas, self).__init__(fig)


class ExportDialog(QDialog):
    def __init__(self, parent=None):
        super(ExportDialog, self).__init__(parent)
        self.parent = parent
        self.setWindowTitle('Save Energy to b-factor')
        self.current_frame_s = QSpinBox()
        self.current_frame_l = QLabel('Frame:')

        self.output_label = QLabel('Output:')
        self.output_text = QLineEdit('complex_e2bfactor')
        self.output_text.setPlaceholderText('complex_e2bfactor')
        self.save_btn = QPushButton('Save')
        self.save_btn.clicked.connect(self.save)
        self.close_btn = QPushButton('Close')
        self.close_btn.clicked.connect(self.close)

        self.out_layout = QHBoxLayout()
        self.out_layout.addWidget(self.output_label)
        self.out_layout.addWidget(self.output_text)

        self.frame_layout = QHBoxLayout()
        self.frame_layout.addWidget(self.current_frame_l)
        self.frame_layout.addWidget(self.current_frame_s)
        # self.frame_layout.addWidget(self.interval_l, alignment=Qt.AlignRight)
        # self.frame_layout.addWidget(self.interval_s)

        self.btn_layout = QHBoxLayout()
        self.btn_layout.addStretch(1)
        self.btn_layout.addWidget(self.save_btn, alignment=Qt.AlignRight)
        self.btn_layout.addWidget(self.close_btn, alignment=Qt.AlignRight)

        self.layout = QVBoxLayout(self)
        self.layout.addLayout(self.out_layout)
        # self.layout.addWidget(self.location_btn)
        self.layout.addLayout(self.frame_layout)
        self.layout.addLayout(self.btn_layout)

        self.getdata()

    def getdata(self):
        ext_prog = find_progs(self.parent.app.INPUT)
        self.ntraj = Trajectory(self.parent.app.FILES.complex_prmtop, self.parent.app.FILES.complex_trajs,
                                ext_prog['cpptraj'])
        self.ntraj.Setup(1, 999999, 1)
        last = self.parent.app.INPUT['startframe'] + (self.ntraj.processed_frames - 1) * self.parent.app.INPUT[
            'interval']
        self.current_frame_s.setRange(self.parent.app.INPUT['startframe'], last)
        self.current_frame_s.setSingleStep(self.parent.app.INPUT['interval'])
        self.current_frame_s.setValue(self.parent.app.INPUT['startframe'])

    @pyqtSlot()
    def save(self):
        self.ntraj.Outtraj('_temp_.pdb', frames=str(self.current_frame_s.value()),
                           filetype='pdb')
        self.ntraj.Run('cpptraj.out')
        pdb = PDB()
        pdb.parse('_temp_.pdb')
        with open(str(self.output_text.text()) + '.pdb', 'w') as pdbout:
            for atm in pdb.allAtoms:
                if atm['id'] == 'TER':
                    pdbout.write(pdb.getOutLine(atm, ter=True))
                else:
                    if atm['resnum'] in self.parent.decomp and atm['resname'] in std_aa:
                        atm['b_factor'] = self.parent.decomp[atm['resnum']]
                    else:
                        atm['b_factor'] = 0.0
                    pdbout.write(pdb.getOutLine(atm, ter=False))
        # os.remove('_temp_.pdb')
        self.parent.statusbar.showMessage('Saving {} file... Done'.format(self.output_text.text() + '.pdb'))
        self.close()


class ExportDialogCSV(QDialog):
    def __init__(self, parent=None):
        super(ExportDialogCSV, self).__init__(parent)
        self.parent = parent
        self.setWindowTitle('Export Energy to CSV')

        self.output_label = QLabel('Output:')
        self.output_text = QLineEdit('TOTAL_ENERGY')
        self.output_text.setPlaceholderText('TOTAL_ENERGY')
        self.save_btn = QPushButton('Save')
        self.save_btn.clicked.connect(self.save)
        self.close_btn = QPushButton('Close')
        self.close_btn.clicked.connect(self.close)

        self.out_layout = QHBoxLayout()
        self.out_layout.addWidget(self.output_label)
        self.out_layout.addWidget(self.output_text)

        self.btn_layout = QHBoxLayout()
        self.btn_layout.addStretch(1)
        self.btn_layout.addWidget(self.save_btn, alignment=Qt.AlignRight)
        self.btn_layout.addWidget(self.close_btn, alignment=Qt.AlignRight)

        self.layout = QVBoxLayout(self)
        self.layout.addLayout(self.out_layout)
        self.layout.addLayout(self.btn_layout)

    @pyqtSlot()
    def save(self):
        with open(str(self.output_text.text()) + '.csv', 'w') as out_file:
            if self.parent.gb_data:
                out_file.write('GB data\n')
                out_file.write(','.join(['FRAME'] + [x[0].upper() for x in self.parent.gb_data]) + '\n')
                self.parent.writeData(out_file, self.parent.gb_data)
            if self.parent.mut_gb_data:
                out_file.write('Mutant GB data\n')
                out_file.write(','.join(['FRAME'] + [x[0].upper() for x in self.parent.mut_gb_data]) + '\n')
                self.parent.writeData(out_file, self.parent.mut_gb_data)
            if self.parent.pb_data:
                out_file.write('PB data\n')
                out_file.write(','.join(['FRAME'] + [x[0].upper() for x in self.parent.pb_data]) + '\n')
                self.parent.writeData(out_file, self.parent.pb_data)
            if self.parent.mut_pb_data:
                out_file.write('Mutant PB data\n')
                out_file.write(','.join(['FRAME'] + [x[0].upper() for x in self.parent.mut_pb_data]) + '\n')
                self.parent.writeData(out_file, self.parent.mut_pb_data)
        self.parent.statusbar.showMessage('Exporting PB/GB energy to csv file... Done.')
        self.close()


class GMX_MMPBSA_ANA(QMainWindow):
    def __init__(self, info_file=None):
        super(GMX_MMPBSA_ANA, self).__init__()

        self.data = None
        self.mutant_data = None
        self.app = None
        self.infofile = Path(info_file)

        self.gb_data = None
        self.mut_gb_data = None
        self.pb_data = None
        self.mut_pb_data = None
        self.decomp = {}
        self.mdi = QMdiArea(self)
        self.setCentralWidget(self.mdi)

        self.menubar = self.menuBar()
        # self.opendirAct = QAction("Open gmx_MMPBSA Dir...", self)
        # self.opendirAct.triggered.connect(self.getInfoFile)
        self.fileMenu = self.menuBar().addMenu("&File")
        # self.fileMenu.addAction(self.opendirAct)
        self.fileMenu.addAction('Close All..', self.mdi.closeAllSubWindows)
        self.viewMenu = self.menuBar().addMenu("&View")
        self.viewMenu.addAction('Tile SubWindows', self.mdi.tileSubWindows)
        self.viewMenu.addAction('Cascade SubWindows', self.mdi.cascadeSubWindows)
        self.statusbar = self.statusBar()

        self.treeDockWidget = QDockWidget(self)
        self.treeDockWidget.setWindowTitle('Data')
        self.addDockWidget(Qt.RightDockWidgetArea, self.treeDockWidget)
        self.optionDockWidget = QDockWidget(self)
        self.optionDockWidget.setWindowTitle('Options')
        self.addDockWidget(Qt.RightDockWidgetArea, self.optionDockWidget)

        self.getData()
        if 'gb' in self.data or 'pb' in self.data:
            self.fileMenu.addAction('Export GB/PB energy (csv)', self.exportCSV)
        if 'decomp' in self.data:
            self.fileMenu.addAction('Energy to bfactor', self.energy2bfactor)

        self.exportpdb = ExportDialog(self)
        self.exportcsv = ExportDialogCSV(self)

    def getInfoFile(self):
        info_file, _ = QFileDialog.getOpenFileName(self, "Output Directory", QDir.currentPath(), '*_info',
                                                   options=QFileDialog.DontResolveSymlinks)
        if info_file:
            self.infofile = Path(info_file)
            self.getData()

    def getData(self):
        os.chdir(self.infofile.parent)
        self.data, self.app = API.load_gmxmmpbsa_info(self.infofile.as_posix())
        self.makeTree()

    def showdata(self, item: CustomItem, col):
        if col == 1:
            s = item.subwindows[1]
            if item.checkState(col) == Qt.Checked:
                if s:  # check if any subwindow has been store
                    if not s.isVisible():
                        s.show()
                else:

                    sub = Charts()
                    sub.setObjectName(item.dataperframe['name'])
                    sub.item = item
                    sub.col = col
                    item.subwindows[col] = sub

                    x = item.dataperframe['frames']
                    x = np.array(x)
                    y = item.dataperframe['data']

                    sub.setMinimumSize(500, 300)
                    mainwidgetmdi = QMainWindow()

                    self._main = QWidget()
                    mainwidgetmdi.setCentralWidget(self._main)
                    layout = QVBoxLayout(self._main)
                    sc = MplCanvas(self, width=5, height=4, dpi=100)
                    sc.axes.set_title(item.dataperframe['name'])
                    sc.axes.plot(x, y, color='black', linewidth=0.5)
                    sc.axes.set_xlabel(item.dataperframe['xaxis'])
                    sc.axes.set_ylabel(item.dataperframe['yaxis'])

                    # Create toolbar, passing canvas as first parament, parent (self, the MainWindow) as second.
                    toolbar = NavigationToolbar(sc, self)
                    layout.addWidget(sc)
                    layout.addWidget(toolbar)

                    sub.setWidget(mainwidgetmdi)
                    sub.setWindowTitle(item.dataperframe['name'])
                    self.mdi.addSubWindow(sub)
                    sub.show()
                    sc.draw()
                    sc.figure.tight_layout()
                    sc.draw()
            else:
                if s:  # check if any subwindow has been store
                    if s.isVisible():
                        s.close()

        elif col == 2:
            s = item.subwindows[2]
            if item.checkState(col) == Qt.Checked:
                if s:  # check if any subwindow has been store
                    if not s.isVisible():
                        s.show()
                else:
                    sub = Charts()
                    sub.setObjectName(item.datamean['name'])
                    sub.item = item
                    sub.col = col
                    item.subwindows[col] = sub

                    y = item.datamean['data']

                    sub.setMinimumSize(500, 300)
                    mainwidgetmdi = QMainWindow()

                    self._main = QWidget()
                    mainwidgetmdi.setCentralWidget(self._main)
                    layout = QVBoxLayout(self._main)
                    sc = MplCanvas(self, width=5, height=4, dpi=100)
                    sc.axes.set_title(item.datamean['name'])
                    data_len = len(item.datamean['xaxis'])
                    sc.axes.set_xlim(0, data_len + 1)
                    sc.axes.set_xticks([x for x in range(1, data_len + 1)])
                    sc.axes.set_xticklabels(item.datamean['xaxis'])
                    sc.axes.set_ylabel(item.datamean['yaxis'])
                    for label in sc.axes.get_xticklabels():
                        label.set_rotation(40)
                        label.set_horizontalalignment('right')

                    bottom = 0
                    for i in range(0, data_len):
                        sc.axes.bar(i + 1, y[i])
                        bottom += y[i]

                    # Create toolbar, passing canvas as first parament, parent (self, the MainWindow) as second.
                    toolbar = NavigationToolbar(sc, self)
                    layout.addWidget(sc)
                    layout.addWidget(toolbar)

                    sub.setWidget(mainwidgetmdi)
                    sub.setWindowTitle(item.datamean['name'])
                    self.mdi.addSubWindow(sub)
                    sub.show()
                    sc.draw()
                    sc.figure.tight_layout()
                    sc.draw()
            else:
                if s:  # check if any subwindow has been store
                    if s.isVisible():
                        s.close()

    def writeData(self, outfile, data):
        for i in range(len(self.frames)):
            outfile.write(','.join([str(self.frames[i])] + [str(x[1][i]) for x in data]) + '\n')

    def energy2bfactor(self):
        self.exportpdb.show()

    def exportCSV(self):
        self.exportcsv.show()

    def makeItems(self, data, topItem, mutant=False):
        mut_pre = ''
        if mutant:
            mut_pre = 'Mutant '

        for level in data:
            item = CustomItem([level.upper()])
            topItem.addChild(item)
            dh = None
            if level in ['gb', 'pb', 'rism gf', 'rism std']:
                cd = []
                dat = []
                for level1 in data[level]['delta']:
                    item1 = CustomItem([str(level1).upper()])
                    item1.setCheckState(1, Qt.Unchecked)
                    end = len(data[level]['delta'][level1]) * self.app.INPUT['interval'] + self.app.INPUT['startframe']
                    self.frames = [x for x in range(self.app.INPUT['startframe'], end, self.app.INPUT['interval'])]
                    item1.dataperframe = {
                        'name': mut_pre + '{} {} Energy (Per-Frame)'.format(level.upper(), level1.upper()),
                        'xaxis': 'frames', 'yaxis': 'Energy (kcal/mol)', 'frames': self.frames,
                        'data': data[level]['delta'][level1]}
                    cd.append([level1, data[level]['delta'][level1].mean()])
                    dat.append([level1, data[level]['delta'][level1]])
                    if level1 == 'DELTA TOTAL':
                        item1.setCheckState(2, Qt.Unchecked)
                        item1.datamean = {
                            'name': mut_pre + '{} {} Energy (Mean)'.format(level.upper(), level1.upper()),
                            'xaxis': [x[0].upper() for x in cd], 'yaxis': 'Energy (kcal/mol)',
                            'data': np.array([x[1] for x in cd])}
                        dh = data[level]['delta'][level1].mean()
                    item.addChild(item1)
                    if level1 == 'DELTA TOTAL' and 'nmode' in data:
                        itemd = CustomItem(['DELTA G Binding (NM)'])
                        itemd.setCheckState(2, Qt.Unchecked)
                        entropy = data['nmode']['delta']['Total'][0]
                        itemd.datamean = {
                            'name': mut_pre + '{} Binding Energy (with Entropy (nmode))'.format(level.upper()),
                            'xaxis': ['ΔH', '-TΔS', 'ΔG'], 'yaxis': 'Energy (kcal/mol)',
                            'data': np.array([dh, entropy * -1, dh - entropy])}
                        item.addChild(itemd)
                    if level1 == 'DELTA TOTAL' and 'qh' in data:
                        itemd = CustomItem(['DELTA G Binding (QH)'])
                        itemd.setCheckState(2, Qt.Unchecked)
                        entropy = data['qh']['delta']['Total'][0]

                        itemd.datamean = {
                            'name': mut_pre + '{} Binding Energy (with Entropy (qh))'.format(level.upper()),
                            'xaxis': ['ΔH', '-TΔS', 'ΔG'], 'yaxis': 'Energy (kcal/mol)',
                            'data': np.array([dh, entropy * -1, dh - entropy])}
                        item.addChild(itemd)
                    if level1 == '-TDS':
                        itemd = CustomItem(['DELTA G Binding (IE)'])
                        itemd.setCheckState(2, Qt.Unchecked)
                        s = 1
                        if len(data[level]['delta']['-TDS']) * self.app.INPUT['entropy_seg'] / 100 > 1:
                            s = ceil(len(data[level]['delta']['-TDS']) * (1 - self.app.INPUT['entropy_seg'] / 100))
                        ts = data[level]['delta']['-TDS'][s:].mean()
                        itemd.datamean = {
                            'name': mut_pre + '{} Total Energy (with Entropy)'.format(level.upper()),
                            'xaxis': ['ΔH', '-TΔS', 'ΔG'], 'yaxis': 'Energy (kcal/mol)',
                            'data': np.array([dh, ts, dh + ts])}
                        item.addChild(itemd)
                # To export data in csv
                if level == 'gb':
                    if mutant:
                        self.mut_gb_data = dat
                    else:
                        self.gb_data = dat
                else:
                    if mutant:
                        self.mut_pb_data = dat
                    else:
                        self.pb_data = dat

            elif level == 'decomp':
                for level1 in data[level]:
                    item1 = CustomItem([str(level1).upper()])
                    item.addChild(item1)
                    for level2 in data[level][level1]['delta']:
                        item2 = CustomItem([str(level2).upper()])
                        item1.addChild(item2)
                        item2.setCheckState(1, Qt.Unchecked)
                        item2.setCheckState(2, Qt.Unchecked)
                        lvl2_data = np.zeros(int(self.app.numframes / self.app.INPUT['interval']))
                        lvl2_meandata = []
                        for level3 in data[level][level1]['delta'][level2]:
                            item3 = CustomItem([str(level3).upper()])
                            item2.addChild(item3)
                            lvl3_data = np.zeros(int(self.app.numframes / self.app.INPUT['interval']))
                            lvl3_meandata = []
                            lvl4_data = []
                            lvl4_meandata = []
                            for level4 in data[level][level1]['delta'][level2][level3]:
                                item4 = CustomItem([str(level4).upper()])
                                item3.addChild(item4)
                                lvl5_meandata = []
                                if self.app.INPUT['idecomp'] in [3, 4]:
                                    for level5 in data[level][level1]['delta'][level2][level3][level4]:
                                        item5 = CustomItem([str(level5).upper()])
                                        item4.addChild(item5)
                                        item5.setCheckState(1, Qt.Unchecked)
                                        item5.dataperframe = {
                                            'name': mut_pre + 'Docomposition ({}) {} Residue {} Pair {} {} Energy '
                                                              '(Per-Frame)'.format(level1.upper(), level2.upper(),
                                                                                   level3, level4, level5.upper()),
                                            'xaxis': 'frames', 'yaxis': 'Energy (kcal/mol)', 'frames': self.frames,
                                            'data': data[level][level1]['delta'][level2][level3][level4][level5]}
                                        lvl5_meandata.append(
                                            [level5,
                                             data[level][level1]['delta'][level2][level3][level4][level5].mean()])
                                        if level5 == 'tot':
                                            item5.setCheckState(2, Qt.Unchecked)
                                            lvl3_data += data[level][level1]['delta'][level2][level3][level4][level5]
                                            lvl3_meandata.append(
                                                [level4,
                                                 data[level][level1]['delta'][level2][level3][level4][level5].mean()])

                                            item5.datamean = {
                                                'name': mut_pre + 'Decomposition ({}) {} Residue {} Pair {} '
                                                                  'Total Energy (Mean)'.format(level1.upper(),
                                                                                               level2.upper(),
                                                                                               level3, level4),
                                                'xaxis': [x[0].upper() for x in lvl5_meandata],
                                                'yaxis': 'Energy (kcal/mol)',
                                                'data': np.array([x[1] for x in lvl5_meandata])}
                                else:
                                    lvl3_data += data[level][level1]['delta'][level2][level3][level4]
                                    lvl3_meandata.append(
                                        [level4, data[level][level1]['delta'][level2][level3][level4].mean()])
                                    item4.setCheckState(1, Qt.Unchecked)
                                    item4.dataperframe = {
                                        'name': mut_pre + 'Docomposition ({}) {} Residue {} {} Energy '
                                                          '(Per-Frame)'.format(level1.upper(), level2.upper(), level3,
                                                                               level4.upper()),
                                        'xaxis': 'frames', 'yaxis': 'Energy (kcal/mol)', 'frames': self.frames,
                                        'data': data[level][level1]['delta'][level2][level3][level4]}
                                    lvl4_meandata.append(
                                        [level4, data[level][level1]['delta'][level2][level3][level4].mean()])
                                    if level4 == 'tot':
                                        item4.setCheckState(2, Qt.Unchecked)
                                        item4.datamean = {
                                            'name': mut_pre + 'Decomposition ({}) {} Residue {} Total Energy ('
                                                              'Mean)'.format(level1.upper(), level2.upper(), level3),
                                            'xaxis': [x[0].upper() for x in lvl4_meandata],
                                            'yaxis': 'Energy (kcal/mol)',
                                            'data': np.array([x[1] for x in lvl4_meandata]), }

                            if 'tot' in data[level][level1]['delta'][level2][level3]:  # Per-residue
                                lvl2_data += lvl3_data
                                lvl2_meandata.append([level3, lvl3_data.mean()])
                                if level2 == 'TDC':
                                    self.decomp[level3] = lvl3_data.mean()
                            else:  # Per-wise
                                item3.setCheckState(1, Qt.Unchecked)
                                item3.dataperframe = {
                                    'name': mut_pre + 'Decomposition ({})  {} Residue {} Energy (Per-Frame)'.format(
                                        level1.upper(), level2.upper(), level3), 'xaxis': 'frames',
                                    'yaxis': 'Energy (kcal/mol)', 'frames': self.frames,
                                    'data': lvl3_data}
                                if level2 == 'TDC':
                                    self.decomp[level3] = sum([x[1] for x in lvl3_meandata])
                                lvl3_meandata.extend([['TOTAL', sum([x[1] for x in lvl3_meandata])]])
                                lvl2_data += lvl3_data
                                lvl2_meandata.append([level3, lvl3_data.mean()])

                                item3.datamean = {
                                    'name': mut_pre + 'Decomposition ({}) {} Residue {} Energy (Mean) from '
                                                      'Pairs'.format(level1.upper(), level2.upper(), level3),
                                    'xaxis': [x[0] for x in lvl3_meandata], 'yaxis': 'Energy (kcal/mol)',
                                    'data': np.array([x[1] for x in lvl3_meandata])}
                                item3.setCheckState(2, Qt.Unchecked)
                        item2.dataperframe = {
                            'name': mut_pre + 'Decomposition ({})  {} Energy (Per-Frame)'.format(
                                level1.upper(), level2.upper()), 'xaxis': 'frames', 'yaxis': 'Energy (kcal/mol)',
                            'frames': self.frames, 'data': lvl2_data}
                        lvl2_meandata.extend([['TOTAL', sum([x[1] for x in lvl2_meandata])]])
                        item2.datamean = {
                            'name': mut_pre + 'Decomposition ({}) {} Energy (Mean)'.format(
                                level1.upper(), level2.upper()), 'xaxis': [x[0] for x in lvl2_meandata],
                            'yaxis': 'Energy (kcal/mol)', 'data': np.array([x[1] for x in lvl2_meandata])}

            elif level in ['nmode', 'qh']:
                item.setCheckState(2, Qt.Unchecked)
                item.datamean = {
                    'name': mut_pre + 'Entropy ({})'.format(level.upper()),
                    'xaxis': [x[0].upper() for x in data[level]['delta'].items()], 'yaxis': 'Energy (kcal/mol)',
                    'data': np.array([x[1][0] for x in data[level]['delta'].items()])}

    def makeTree(self):
        self.treeWidget = QTreeWidget(self)
        self.treeWidget.itemChanged.connect(self.showdata)
        self.treeDockWidget.setWidget(self.treeWidget)
        self.treeWidget.setHeaderLabels(['Labels', 'Per-Frame', 'Mean'])

        self.normalitem = CustomItem(['Normal System'])
        self.normalitem.setExpanded(True)
        self.treeWidget.addTopLevelItem(self.normalitem)
        self.makeItems(self.data, self.normalitem)
        self.normalitem.setExpanded(True)

        if self.data.mutant:
            self.mutantitem = CustomItem(['Mutant System'])
            self.treeWidget.addTopLevelItem(self.mutantitem)
            self.makeItems(self.data.mutant, self.mutantitem, True)
            self.mutantitem.setExpanded(True)

        self.treeWidget.header().setSectionResizeMode(QHeaderView.ResizeToContents)


def run(infofile):
    app = QApplication(sys.argv)
    app.setApplicationName('gmx_MMPBSA')
    w = GMX_MMPBSA_ANA(infofile)
    w.show()
    sys.exit(app.exec())


if __name__ == '__main__':
    app = QApplication(sys.argv)
    app.setApplicationName('gmx_MMPBSA')

    w = GMX_MMPBSA_ANA('/home/mario/Drive/scripts/gmx_MMPBSA/test/qh/_GMXMMPBSA_info')
    w.show()
    sys.exit(app.exec())
